#!/usr/bin/env python3
import logging
import ntpath
from io import BytesIO
from impacket.smbconnection import SMBConnection
from impacket.smb3structs import FILE_ATTRIBUTE_ARCHIVE, FILE_ATTRIBUTE_COMPRESSED, FILE_ATTRIBUTE_DIRECTORY, FILE_ATTRIBUTE_ENCRYPTED, FILE_ATTRIBUTE_HIDDEN, FILE_ATTRIBUTE_NORMAL, FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, FILE_ATTRIBUTE_OFFLINE, FILE_ATTRIBUTE_READONLY, FILE_ATTRIBUTE_REPARSE_POINT, FILE_ATTRIBUTE_SPARSE_FILE, FILE_ATTRIBUTE_SYSTEM, FILE_ATTRIBUTE_TEMPORARY, FILE_ATTRIBUTE_INTEGRITY_STREAM, FILE_ATTRIBUTE_NO_SCRUB_DATA
from impacket.dcerpc.v5 import transport, srvs
from impacket.dcerpc.v5.dtypes import OWNER_SECURITY_INFORMATION, GROUP_SECURITY_INFORMATION, DACL_SECURITY_INFORMATION, SACL_SECURITY_INFORMATION

from powerview.utils.accesscontrol import AccessControl, SIMPLE_PERMISSIONS

class SMBClient:
    def __init__(self, client):
        self.client = client

    def shares(self):
        if self.client is None:
            logging.error("[SMBClient: shares] Not logged in")
            return
        
        return self.client.listShares()

    def add_share(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: add_share] Not logged in")
            return
        
        rpctransport = transport.SMBTransport(self.client.getRemoteName(), self.client.getRemoteHost(), filename=r'\srvsvc',
                                              smb_connection=self.client)
        dce = rpctransport.get_dce_rpc()
        dce.connect()
        dce.bind(srvs.MSRPC_UUID_SRVS)

        # Use SHARE_INFO_2 instead of 502 for simpler structure
        info_2 = srvs.SHARE_INFO_2()
        info_2['shi2_netname'] = share.replace('/', '\\').replace('\\\\', '\\') + '\x00'
        info_2['shi2_type'] = 0  # STYPE_DISKTREE
        info_2['shi2_remark'] = 'Created by PowerView\x00'
        info_2['shi2_permissions'] = 0x00000000
        info_2['shi2_max_uses'] = 0xFFFFFFFF  # No limit
        info_2['shi2_current_uses'] = 0
        info_2['shi2_path'] = path + '\x00'
        info_2['shi2_passwd'] = '\x00'
        
        try:
            resp = srvs.hNetrShareAdd(
                dce,
                2,  # Use level 2
                info_2
            )
            if resp['ErrorCode'] != 0:
                error_msg = f"Error code: 0x{resp['ErrorCode']:x}"
                logging.error(f"[SMBClient: add_share] Error adding share: {error_msg}")
                raise Exception(f"[SMBClient: add_share] Error adding share: {error_msg}")
            else:
                logging.debug(f"[SMBClient: add_share] Successfully added share: {share}")
                return True
        except Exception as e:
            logging.error(f"[SMBClient: add_share] Error adding share: {e}")
            return False

    def delete_share(self, share):
        if self.client is None:
            logging.error("[SMBClient: delete_share] Not logged in")
            return
        
        rpctransport = transport.SMBTransport(self.client.getRemoteName(), self.client.getRemoteHost(), filename=r'\srvsvc',
                                              smb_connection=self.client)
        dce = rpctransport.get_dce_rpc()
        dce.connect()
        dce.bind(srvs.MSRPC_UUID_SRVS)
        
        try:
            resp = srvs.hNetrShareDel(dce, share.replace('/', '\\').replace('\\\\', '\\') + '\x00')
            if resp['ErrorCode'] != 0:
                error_msg = f"Error code: 0x{resp['ErrorCode']:x}"
                logging.error(f"[SMBClient: delete_share] Error deleting share: {error_msg}")
                raise Exception(f"[SMBClient: delete_share] Error deleting share: {error_msg}")
            else:
                logging.debug(f"[SMBClient: delete_share] Successfully deleted share: {share}")
                return True
        except Exception as e:
            logging.error(f"[SMBClient: delete_share] Error deleting share: {e}")
            return False

    def share_info(self, share):
        if self.client is None:
            logging.error("[SMBClient: share_info] Not logged in")
            return
        
        rpctransport = transport.SMBTransport(self.client.getRemoteName(), self.client.getRemoteHost(), filename=r'\srvsvc',
                                              smb_connection=self.client)
        dce = rpctransport.get_dce_rpc()
        dce.connect()
        dce.bind(srvs.MSRPC_UUID_SRVS)
        
        share_info = {
            'name': None,
            'type': None,
            'remark': None,
            'path': None,
            'permissions': None,
            'max_uses': None,
            'current_uses': None,
            'passwd': None,
            'servername': None,
            'reserved': None,
            'sd_info': None
        }

        try:
            base_info = srvs.hNetrShareGetInfo(dce, share + '\x00', 1)
            share_info['name'] = base_info['InfoStruct']['ShareInfo1']['shi1_netname'][:-1]
            share_info['type'] = base_info['InfoStruct']['ShareInfo1']['shi1_type']
            share_info['remark'] = base_info['InfoStruct']['ShareInfo1']['shi1_remark'][:-1]
        except Exception as e:
            logging.error(f"[SMBClient: share_info] Error getting share info via NetrShareGetInfo(Level 1): {e}")

        try:
            resp = srvs.hNetrShareGetInfo(dce, share + '\x00', 502)
            share_info['path'] = resp['InfoStruct']['ShareInfo502']['shi502_path'][:-1]
            share_info['permissions'] = resp['InfoStruct']['ShareInfo502']['shi502_permissions']
            share_info['max_uses'] = resp['InfoStruct']['ShareInfo502']['shi502_max_uses']
            share_info['current_uses'] = resp['InfoStruct']['ShareInfo502']['shi502_current_uses']
            share_info['passwd'] = resp['InfoStruct']['ShareInfo502']['shi502_passwd']
            share_info['reserved'] = resp['InfoStruct']['ShareInfo502']['shi502_reserved']
            secDesc = resp['InfoStruct']['ShareInfo502']['shi502_security_descriptor']
            if secDesc and len(secDesc) > 0:
                share_info['sd_info'] = AccessControl.parse_sd(secDesc)
        except Exception as e:
            logging.error(f"[SMBClient: share_info] Error getting share info via NetrShareGetInfo(Level 502): {e}")
        
        return share_info

    def set_share_security(self, share, sid, mask='fullcontrol', ace_type='allow'):
        if self.client is None:
            logging.error("[SMBClient: set_share_security] Not logged in")
            return

        # convert mask to integer
        mask = mask.lower()
        if mask == 'fullcontrol':
            mask = SIMPLE_PERMISSIONS.FullControl.value
        elif mask == 'modify':
            mask = SIMPLE_PERMISSIONS.Modify.value
        elif mask == 'readandexecute':
            mask = SIMPLE_PERMISSIONS.ReadAndExecute.value
        elif mask == 'readandwrite':
            mask = SIMPLE_PERMISSIONS.ReadAndWrite.value
        elif mask == 'read':
            mask = SIMPLE_PERMISSIONS.Read.value
        elif mask == 'write':
            mask = SIMPLE_PERMISSIONS.Write.value
        else:
            raise Exception(f"[SMBClient: set_share_security] Invalid mask: {mask}")
        
        try:
            rpctransport = transport.SMBTransport(self.client.getRemoteName(), self.client.getRemoteHost(), filename=r'\srvsvc',
                                              smb_connection=self.client)
            dce = rpctransport.get_dce_rpc()
            dce.connect()
            dce.bind(srvs.MSRPC_UUID_SRVS)
            
            logging.debug(f"[SMBClient: set_share_security] Getting share security")
            resp = srvs.hNetrShareGetInfo(dce, share + '\x00', 502)
            secDesc = resp['InfoStruct']['ShareInfo502']['shi502_security_descriptor']

            if ace_type == 'allow':
                security_descriptor = AccessControl.add_allow_ace(
                    secDesc,
                    sid,
                    mask
                )
            elif ace_type == 'deny':
                security_descriptor = AccessControl.add_deny_ace(
                    secDesc,
                    sid,
                    mask
                )
            else:
                raise Exception(f"[SMBClient: set_share_security] Invalid ace_type: {ace_type}")

            logging.debug(f"[SMBClient: set_share_security] Setting share security")
            info_1501 = srvs.SHARE_INFO_1501()
            info_1501['shi1501_security_descriptor'] = security_descriptor

            resp = srvs.hNetrShareSetInfo(dce, share + '\x00', 1501, info_1501)
            if resp['ErrorCode'] != 0:
                raise Exception(f"[SMBClient: set_share_security] Error setting share security")
            else:
                logging.debug(f"[SMBClient: set_share_security] Successfully set share security")
        except Exception as e:
            logging.error(f"[SMBClient: set_share_security] Error setting share security: {e}")
            return False
        
        return True

    def remove_share_security(self, share, sid, mask=None, ace_type=None):
        if self.client is None:
            logging.error("[SMBClient: remove_share_security] Not logged in")
            return
        
        try:
            rpctransport = transport.SMBTransport(self.client.getRemoteName(), self.client.getRemoteHost(), filename=r'\srvsvc',
                                              smb_connection=self.client)
            dce = rpctransport.get_dce_rpc()
            dce.connect()
            dce.bind(srvs.MSRPC_UUID_SRVS)
            
            logging.debug(f"[SMBClient: remove_share_security] Getting share security")
            resp = srvs.hNetrShareGetInfo(dce, share + '\x00', 502)
            secDesc = resp['InfoStruct']['ShareInfo502']['shi502_security_descriptor']

            mask_value = None
            if mask:
                if mask == 'fullcontrol':
                    mask_value = SIMPLE_PERMISSIONS.FullControl.value
                elif mask == 'modify':
                    mask_value = SIMPLE_PERMISSIONS.Modify.value
                elif mask == 'readandexecute':
                    mask_value = SIMPLE_PERMISSIONS.ReadAndExecute.value
                elif mask == 'readandwrite':
                    mask_value = SIMPLE_PERMISSIONS.ReadAndWrite.value
                elif mask == 'read':
                    mask_value = SIMPLE_PERMISSIONS.Read.value
                elif mask == 'write':
                    mask_value = SIMPLE_PERMISSIONS.Write.value
                else:
                    raise Exception(f"[SMBClient: remove_share_security] Invalid mask: {mask}")

            ace_type_value = None
            if ace_type:
                if ace_type == 'allow':
                    from impacket.ldap import ldaptypes
                    ace_type_value = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
                elif ace_type == 'deny':
                    from impacket.ldap import ldaptypes
                    ace_type_value = ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE
                else:
                    raise Exception(f"[SMBClient: remove_share_security] Invalid ace_type: {ace_type}")

            security_descriptor, removed_count = AccessControl.remove_ace(
                secDesc,
                sid,
                mask_value,
                ace_type_value
            )

            if removed_count == 0:
                logging.warning(f"[SMBClient: remove_share_security] No matching ACEs found to remove")
                return False

            logging.debug(f"[SMBClient: remove_share_security] Setting share security")
            info_1501 = srvs.SHARE_INFO_1501()
            info_1501['shi1501_security_descriptor'] = security_descriptor

            resp = srvs.hNetrShareSetInfo(dce, share + '\x00', 1501, info_1501)
            if resp['ErrorCode'] != 0:
                raise Exception(f"[SMBClient: remove_share_security] Error setting share security")
            else:
                logging.debug(f"[SMBClient: remove_share_security] Successfully removed {removed_count} ACE(s)")
            return True
        except Exception as e:
            logging.error(f"[SMBClient: remove_share_security] Error removing share security: {e}")
            return False

    def ls(self, share, path=''):
        if self.client is None:
            logging.error("[SMBClient: ls] Not logged in")
            return
        
        path = path.replace('/', '\\')
        path = ntpath.join(path, '*')
        
        return self.client.listPath(share, ntpath.normpath(path))

    def mv(self, share, source, destination):
        if self.client is None:
            logging.error("[SMBClient: mv] Not logged in")
            return

        source = source.replace('/', '\\')
        
        self.client.rename(share, source, destination)
    
    def get(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: get] Not logged in")
            return
        
        path = path.replace('/', '\\')
        fh = BytesIO()
        try:
            self.client.getFile(share, ntpath.normpath(path), fh.write)
            return fh.getvalue()
        except:
            raise
        finally:
            fh.close()

    def put(self, share, remote_path, local_path):
        if self.client is None:
            logging.error("[SMBClient: put] Not logged in")
            return
        
        try:
            with open(local_path, 'rb') as fh:
                # Normalize the remote path for the target OS (Windows)
                normalized_remote_path = remote_path.replace('/', '\\')
                final_remote_path = ntpath.normpath(normalized_remote_path)
                logging.debug(f"[SMBClient: put] Uploading local '{local_path}' to share '{share}' path '{final_remote_path}'")
                self.client.putFile(share, final_remote_path, fh.read)
        except FileNotFoundError:
            logging.error(f"[SMBClient: put] Local file not found: {local_path}")
            raise Exception(f"Local file not found: {local_path}")
        except Exception as e:
            logging.error(f"[SMBClient: put] Error during upload to {share}\\{remote_path}: {e}")
            raise e

    def cat(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: cat] Not logged in")
            return
        
        path = path.replace('/', '\\')
        fh = BytesIO()
        try:
            self.client.getFile(share, ntpath.normpath(path), fh.write)
            return fh.getvalue()
        except:
            raise
        finally:
            fh.close()

    def rm(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: rm] Not logged in")
            return
        
        self.client.deleteFile(share, path)

    def rmdir(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: rmdir] Not logged in")
            return

        path = path.replace('/', '\\')
        self.client.deleteDirectory(share, path)

    def mkdir(self, share, path):
        if self.client is None:
            logging.error("[SMBClient: mkdir] Not logged in")
            return
        
        path = path.replace('/', '\\')
        self.client.createDirectory(share, path)

    def get_file_info(self, share, path):
        """Get detailed information about a file or directory."""
        if self.client is None:
            logging.error("[SMBClient: get_file_info] Not logged in")
            return None
        
        path = path.replace('/', '\\')
        normalized_path = ntpath.normpath(path)
        
        try:
            # For files, we need the file itself
            file_obj = None
            is_dir = False
            
            # Check if this is a directory by attempting to list it
            try:
                parent_dir = ntpath.dirname(normalized_path)
                file_name = ntpath.basename(normalized_path)
                
                # If path is root or has no parent, adjust accordingly
                if not parent_dir:
                    parent_path = '*'
                    items = self.client.listPath(share, parent_path)
                    for item in items:
                        if item.get_longname() == file_name:
                            file_obj = item
                            break
                else:
                    search_path = ntpath.join(parent_dir, '*')
                    items = self.client.listPath(share, search_path)
                    for item in items:
                        if item.get_longname() == file_name:
                            file_obj = item
                            break
                
                # Try to check if it's a directory
                if file_obj and file_obj.is_directory():
                    is_dir = True
                
            except Exception as e:
                logging.debug(f"[SMBClient: get_file_info] Error checking if path is directory: {e}")
                # If we can't determine if it's a directory, try to get the file directly
                pass
            
            info = {}
            
            if file_obj:
                # Basic file information from FileInfo object
                info = {
                    'name': file_obj.get_longname(),
                    'short_name': file_obj.get_shortname(),
                    'size': file_obj.get_filesize(),
                    'is_directory': is_dir,
                    'created': str(file_obj.get_ctime()),
                    'modified': str(file_obj.get_mtime()),
                    'accessed': str(file_obj.get_atime()),
                    'attributes': file_obj.get_attributes()
                }
                
                # Add attribute flags interpretation
                attr_flags = []
                attr_value = file_obj.get_attributes()
                
                # Standard file attribute flags
                if attr_value & FILE_ATTRIBUTE_READONLY:
                    attr_flags.append("READ_ONLY")
                if attr_value & FILE_ATTRIBUTE_HIDDEN:
                    attr_flags.append("HIDDEN")
                if attr_value & FILE_ATTRIBUTE_SYSTEM:
                    attr_flags.append("SYSTEM")
                if attr_value & FILE_ATTRIBUTE_DIRECTORY:
                    attr_flags.append("DIRECTORY")
                if attr_value & FILE_ATTRIBUTE_ARCHIVE:
                    attr_flags.append("ARCHIVE")
                if attr_value & FILE_ATTRIBUTE_NORMAL:
                    attr_flags.append("NORMAL")
                if attr_value & FILE_ATTRIBUTE_TEMPORARY:
                    attr_flags.append("TEMPORARY")
                if attr_value & FILE_ATTRIBUTE_SPARSE_FILE:
                    attr_flags.append("SPARSE_FILE")
                if attr_value & FILE_ATTRIBUTE_REPARSE_POINT:
                    attr_flags.append("REPARSE_POINT")
                if attr_value & FILE_ATTRIBUTE_COMPRESSED:
                    attr_flags.append("COMPRESSED")
                if attr_value & FILE_ATTRIBUTE_OFFLINE:
                    attr_flags.append("OFFLINE")
                if attr_value & FILE_ATTRIBUTE_NOT_CONTENT_INDEXED:
                    attr_flags.append("NOT_CONTENT_INDEXED")
                if attr_value & FILE_ATTRIBUTE_ENCRYPTED:
                    attr_flags.append("ENCRYPTED")
                
                info['attribute_flags'] = attr_flags

                # get security descriptor of the file
                try:
                    rpctransport = transport.SMBTransport(
                        self.client.getRemoteName(),
                        self.client.getRemoteHost(),
                        filename=r'\srvsvc',
                        smb_connection=self.client
                    )
                    dce = rpctransport.get_dce_rpc()
                    dce.connect()
                    dce.bind(srvs.MSRPC_UUID_SRVS)

                    security_flags = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION

                    if path.startswith('\\'):
                        path = path[1:]  # Remove leading backslash if present
                    path = path.replace('/', '\\')

                    resp = srvs.hNetrpGetFileSecurity(dce, share+'\x00', path+'\x00', security_flags)
                    if resp and len(resp) > 0:
                        info['sd_info'] = AccessControl.parse_sd(resp)
                except Exception as rpc_error:
                    raise Exception(f"[SMBClient: get_file_info] RPC error: {rpc_error}")
            return info
            
        except Exception as e:
            logging.error(f"[SMBClient: get_file_info] Error: {e}")
            raise

    def set_file_security(self, share, path, sid, ace_type='allow', mask='fullcontrol'):
        if self.client is None:
            logging.error("[SMBClient: set_file_security] Not logged in")
            return
        
        path = path.replace('/', '\\')
        try:
            rpctransport = transport.SMBTransport(
                self.client.getRemoteName(),
                self.client.getRemoteHost(),
                filename=r'\srvsvc',
                smb_connection=self.client
            )
            dce = rpctransport.get_dce_rpc()
            dce.connect()
            dce.bind(srvs.MSRPC_UUID_SRVS)

            security_flags = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION

            if path.startswith('\\'):
                path = path[1:]
            path = path.replace('/', '\\')

            secDesc = srvs.hNetrpGetFileSecurity(
                dce, 
                share+'\x00', 
                path+'\x00', 
                security_flags
            )

            mask = mask.lower()
            if mask == 'fullcontrol':
                mask = SIMPLE_PERMISSIONS.FullControl.value
            elif mask == 'modify':
                mask = SIMPLE_PERMISSIONS.Modify.value
            elif mask == 'readandexecute':
                mask = SIMPLE_PERMISSIONS.ReadAndExecute.value
            elif mask == 'readandwrite':
                mask = SIMPLE_PERMISSIONS.ReadAndWrite.value
            elif mask == 'read':
                mask = SIMPLE_PERMISSIONS.Read.value
            elif mask == 'write':
                mask = SIMPLE_PERMISSIONS.Write.value
            else:
                raise Exception(f"[SMBClient: set_file_security] Invalid mask: {mask}")

            if ace_type == 'allow':
                security_descriptor = AccessControl.add_allow_ace(
                    secDesc,
                    sid,
                    mask
                )
            elif ace_type == 'deny':
                security_descriptor = AccessControl.add_deny_ace(
                    secDesc,
                    sid,
                    mask
                )
            else:
                raise Exception(f"[SMBClient: set_file_security] Invalid ace_type: {ace_type}")

            resp = srvs.hNetrpSetFileSecurity(
                dce,
                share+'\x00',
                path+'\x00',
                security_flags,
                security_descriptor
            )
            if resp['ErrorCode'] != 0:
                raise Exception(f"[SMBClient: set_file_security] Error setting file security")
            else:
                logging.debug(f"[SMBClient: set_file_security] Successfully set file security")
            return True
        except Exception as rpc_error:
            raise Exception(f"[SMBClient: set_file_security] RPC error: {rpc_error}")

    def remove_file_security(self, share, path, sid, mask=None, ace_type=None):
        if self.client is None:
            logging.error("[SMBClient: remove_file_security] Not logged in")
            return
        
        path = path.replace('/', '\\')
        try:
            rpctransport = transport.SMBTransport(
                self.client.getRemoteName(),
                self.client.getRemoteHost(),
                filename=r'\srvsvc',
                smb_connection=self.client
            )
            dce = rpctransport.get_dce_rpc()
            dce.connect()
            dce.bind(srvs.MSRPC_UUID_SRVS)

            security_flags = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION

            if path.startswith('\\'):
                path = path[1:]
            path = path.replace('/', '\\')

            secDesc = srvs.hNetrpGetFileSecurity(
                dce, 
                share+'\x00', 
                path+'\x00', 
                security_flags
            )

            mask_value = None
            if mask:
                if mask == 'fullcontrol':
                    mask_value = SIMPLE_PERMISSIONS.FullControl.value
                elif mask == 'modify':
                    mask_value = SIMPLE_PERMISSIONS.Modify.value
                elif mask == 'readandexecute':
                    mask_value = SIMPLE_PERMISSIONS.ReadAndExecute.value
                elif mask == 'readandwrite':
                    mask_value = SIMPLE_PERMISSIONS.ReadAndWrite.value
                elif mask == 'read':
                    mask_value = SIMPLE_PERMISSIONS.Read.value
                elif mask == 'write':
                    mask_value = SIMPLE_PERMISSIONS.Write.value
                else:
                    raise Exception(f"[SMBClient: remove_file_security] Invalid mask: {mask}")

            ace_type_value = None
            if ace_type:
                if ace_type == 'allow':
                    ace_type_value = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
                elif ace_type == 'deny':
                    ace_type_value = ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE
                else:
                    raise Exception(f"[SMBClient: remove_file_security] Invalid ace_type: {ace_type}")

            security_descriptor, removed_count = AccessControl.remove_ace(
                secDesc,
                sid,
                mask_value,
                ace_type_value
            )

            if removed_count == 0:
                logging.warning(f"[SMBClient: remove_file_security] No matching ACEs found to remove")
                return False

            resp = srvs.hNetrpSetFileSecurity(
                dce,
                share+'\x00',
                path+'\x00',
                security_flags,
                security_descriptor
            )
            if resp['ErrorCode'] != 0:
                raise Exception(f"[SMBClient: remove_file_security] Error setting file security")
            else:
                logging.debug(f"[SMBClient: remove_file_security] Successfully removed {removed_count} ACE(s)")
            return True
        except Exception as rpc_error:
            raise Exception(f"[SMBClient: remove_file_security] RPC error: {rpc_error}")


#!/usr/bin/env python3
from os import system, name
import json
from powerview.utils.colors import bcolors

# define our clear function
def clear_screen():

    # for windows
    if name == 'nt':
        _ = system('cls')

    # for mac and linux(here, os.name is 'posix')
    else:
        _ = system('clear')



# https://raw.githubusercontent.com/fortra/impacket/refs/heads/master/examples/GetNPUsers.py
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.univ import noValue

from impacket.krb5 import constants
from impacket.krb5.types import KerberosTime, Principal
from impacket.krb5.asn1 import AS_REQ, KERB_PA_PAC_REQUEST, KRB_ERROR, AS_REP, seq_set, seq_set_iter
from impacket.krb5.kerberosv5 import sendReceive, KerberosError

import logging
import datetime
import random
from binascii import hexlify

class ASREProast:
    def __init__(self, powerview):
        self.powerview = powerview
        self.__requestTGT = self.powerview.use_kerberos
        self.__domain = self.powerview.domain
        self.__kdcIP = self.powerview.dc_ip

    @staticmethod
    def getUnixTime(t):
        t -= 116444736000000000
        t /= 10000000
        return t

    def getTGT(self, userName, requestPAC=True):

        clientName = Principal(userName, type=constants.PrincipalNameType.NT_PRINCIPAL.value)

        asReq = AS_REQ()

        domain = self.__domain.upper()
        serverName = Principal('krbtgt/%s' % domain, type=constants.PrincipalNameType.NT_PRINCIPAL.value)

        pacRequest = KERB_PA_PAC_REQUEST()
        pacRequest['include-pac'] = requestPAC
        encodedPacRequest = encoder.encode(pacRequest)

        asReq['pvno'] = 5
        asReq['msg-type'] = int(constants.ApplicationTagNumbers.AS_REQ.value)

        asReq['padata'] = noValue
        asReq['padata'][0] = noValue
        asReq['padata'][0]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_PAC_REQUEST.value)
        asReq['padata'][0]['padata-value'] = encodedPacRequest

        reqBody = seq_set(asReq, 'req-body')

        opts = list()
        opts.append(constants.KDCOptions.forwardable.value)
        opts.append(constants.KDCOptions.renewable.value)
        opts.append(constants.KDCOptions.proxiable.value)
        reqBody['kdc-options'] = constants.encodeFlags(opts)

        seq_set(reqBody, 'sname', serverName.components_to_asn1)
        seq_set(reqBody, 'cname', clientName.components_to_asn1)

        if domain == '':
            raise Exception('Empty Domain not allowed in Kerberos')

        reqBody['realm'] = domain

        now = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)
        reqBody['till'] = KerberosTime.to_asn1(now)
        reqBody['rtime'] = KerberosTime.to_asn1(now)
        reqBody['nonce'] = random.getrandbits(31)

        supportedCiphers = (int(constants.EncryptionTypes.rc4_hmac.value),)

        seq_set_iter(reqBody, 'etype', supportedCiphers)

        message = encoder.encode(asReq)

        try:
            r = sendReceive(message, domain, self.__kdcIP)
        except KerberosError as e:
            if e.getErrorCode() == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value:
                # RC4 not available, OK, let's ask for newer types
                supportedCiphers = (int(constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value),
                                    int(constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value),)
                seq_set_iter(reqBody, 'etype', supportedCiphers)
                message = encoder.encode(asReq)
                r = sendReceive(message, domain, self.__kdcIP)
            elif self.powerview.args.stack_trace:
                raise e
            else:
                logging.error('[ASREProast] %s' % str(e))
                return

        # This should be the PREAUTH_FAILED packet or the actual TGT if the target principal has the
        # 'Do not require Kerberos preauthentication' set
        try:
            asRep = decoder.decode(r, asn1Spec=KRB_ERROR())[0]
        except:
            # Most of the times we shouldn't be here, is this a TGT?
            asRep = decoder.decode(r, asn1Spec=AS_REP())[0]
        else:
            # The user doesn't have UF_DONT_REQUIRE_PREAUTH set
            raise Exception('[ASREProast] User %s doesn\'t have UF_DONT_REQUIRE_PREAUTH set' % userName)

        # Check what type of encryption is used for the enc-part data
        # This will inform how the hash output needs to be formatted
        if asRep['enc-part']['etype'] == 17 or asRep['enc-part']['etype'] == 18:
            return '$krb5asrep$%d$%s$%s$%s$%s' % (asRep['enc-part']['etype'], clientName, domain,
                                                    hexlify(asRep['enc-part']['cipher'].asOctets()[-12:]).decode(),
                                                    hexlify(asRep['enc-part']['cipher'].asOctets()[:-12]).decode())
        else:
            return '$krb5asrep$%d$%s@%s:%s$%s' % (asRep['enc-part']['etype'], clientName, domain,
                                                    hexlify(asRep['enc-part']['cipher'].asOctets()[:16]).decode(),
                                                    hexlify(asRep['enc-part']['cipher'].asOctets()[16:]).decode())

    def request(self, username):
        try:
            return self.getTGT(username)
        except Exception as e:
            raise Exception('[ASREProast] %s' % str(e))


from ldap3 import SEQUENCE_TYPES, MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE, MODIFY_INCREMENT
from ldap3.protocol.rfc4511 import AttributeDescription, PartialAttribute, Vals, Change, Operation, Changes
from ldap3.protocol.convert import prepare_for_sending, validate_attribute_value

from ..templates import LDAP_PUT_FSTRING

from uuid import uuid4
import sys

change_table = {MODIFY_ADD: 0,  # accepts actual values too
                MODIFY_DELETE: 1,
                MODIFY_REPLACE: 2,
                MODIFY_INCREMENT: 3,
                0: 0,
                1: 1,
                2: 2,
                3: 3}

def modify_operation(fqdn,
                    dn,
                    changes,
                    auto_encode,
                    schema=None,
                    validator=None,
                    check_names=False):
    """CRUD on attribute

        Args:
            client (NMFConnection): connected client
            object_ref (str): DN of object to write attribute on
            fqdn (str): fqdn of the DC
            operation (str): operation to preform on the attribute: <MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE> [MS-WSTIM]: 3.2.4.2.3.1
            attribute (str): attribute type including the namespace
            data_type (str): datatype, <'string', 'base64Base'> [MS-ADDM]: 2.3.4
            value (str): string value for attribute in UTF-8

        Returns:
            str: error
    """
    # changes is a dictionary in the form {'attribute': [(operation, [val1, ...]), ...], ...}
    # operation is 0 (add), 1 (delete), 2 (replace), 3 (increment)
    # increment as per RFC4525
    mRequest: str = ""
    change_list = Changes()
    pos = 0
    for attribute in changes:
        for change_operation in changes[attribute]:
            partial_attribute = PartialAttribute()
            partial_attribute['type'] = AttributeDescription(attribute)
            partial_attribute['vals'] = Vals()
            if isinstance(change_operation[1], SEQUENCE_TYPES):
                for index, value in enumerate(change_operation[1]):
                    partial_attribute['vals'].setComponentByPosition(index, prepare_for_sending(validate_attribute_value(schema, attribute, value, auto_encode, validator, check_names=check_names)))
            else:
                partial_attribute['vals'].setComponentByPosition(0, prepare_for_sending(validate_attribute_value(schema, attribute, change_operation[1], auto_encode, validator, check_names=check_names)))
            change = Change()
            change['operation'] = Operation(change_table[change_operation[0]])
            change['modification'] = partial_attribute

            change_list[pos] = change
            pos += 1

        for idx in range(len(change_list)):
                change = change_list[idx]
                operation = change['operation']
                attribute = str(change['modification']['type'])
                vals = change['modification']['vals']
                
                mRequest += f"""<da:ModifyRequest Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/XPath-Level-1">
            <da:Change Operation="{operation}">
                <da:AttributeType>addata:{attribute}</da:AttributeType>
                <da:AttributeValue>"""
                
                data_type = "string"
                for v in vals:
                    value = str(v)
                    mRequest += f"""
                    <ad:value xsi:type="xsd:{data_type}">{value}</ad:value>"""
                
                mRequest += f"""
                </da:AttributeValue>
            </da:Change>
        </da:ModifyRequest>
        """

    put_vars = {
        "object_ref": dn,
        "uuid": str(uuid4()),
        "fqdn": fqdn,
        "attributes": mRequest,
    }
    return LDAP_PUT_FSTRING.format(**put_vars)

def modify_response_to_dict(response):
    pass



#!/usr/bin/env python3
import os
import logging
from datetime import date
from sys import platform
if platform == "linux" or platform == "linux2":
    import gnureadline as readline
else:
    import readline

DEBUG = 'DEBUG'

class CustomFormatter(logging.Formatter):
    grey = '\033[2;37m'
    green = '\033[92m'
    yellow = '\033[93m'
    red = '\033[91m'
    bold_red = '\x1b[31;1m'
    reset = '\033[0m'

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.grey + self.fmt + self.reset,
            logging.INFO: self.green + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

class HistoryConsole:
    def __init__(self, histfile="~/.powerview/.powerview_history"):
        self.histfile = os.path.expanduser(histfile)
        self.init_history(histfile)

    def init_history(self, histfile=None):
        if not histfile:
            histfile = self.histfile

        try:
            if os.path.exists(histfile):
                readline.read_history_file(histfile)
        except FileNotFoundError:
            logging.error(f"History file not found: {histfile}")
        except IOError as e:
            logging.error(f"Error reading history file {histfile}: {e}")

    def save_history(self, histfile=None):
        if not histfile:
            histfile = self.histfile

        try:
            readline.write_history_file(histfile)
        except IOError as e:
            logging.error(f"Error writing history file {histfile}: {e}")

class LOG:
    def __init__(self, folder_name, root_folder=None):
        if not root_folder:
            if os.name == 'nt':
                self.root_folder = os.path.join(os.environ.get('APPDATA', os.path.expanduser('~')), "powerview")
            else:
                self.root_folder = os.path.join(os.path.expanduser('~'), ".powerview")
        else:
            self.root_folder = root_folder
        
        self.folder_name = folder_name.lower()

        self.logs_folder = os.path.join(self.root_folder, "logs", self.folder_name)
        
        if not os.path.exists(self.logs_folder):
            self.create_folder()

        self.file_name = "%s.log" % date.today()

        print("Logging directory is set to %s" % (self.logs_folder))

        self.history_file = os.path.join(self.logs_folder, ".powerview_history")
        self.history_console = HistoryConsole(self.history_file)

    def save_history(self):
        self.history_console.save_history()

    def create_folder(self, folder=None):
        folder = folder if folder else self.logs_folder
        return os.makedirs(folder, exist_ok=True)
    
    def setup_logger(self, level=logging.INFO):
        if level == DEBUG:
            level == logging.DEBUG
       
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        file_path = os.path.join(self.logs_folder, self.file_name)
        fileh = logging.FileHandler(file_path, 'a')
        formatter = logging.Formatter('[%(asctime)s] %(name)s %(levelname)s %(message)s')
        fileh.setFormatter(formatter)
        logger.addHandler(fileh)

        fmt = '[%(asctime)s] %(message)s'
        stdout_handler = logging.StreamHandler()
        stdout_handler.setLevel(level)
        stdout_handler.setFormatter(CustomFormatter(fmt))

        logger.addHandler(stdout_handler)
        return logger

    def write(self, file_name, text):
        success = False

        if not os.path.exists(self.root_folder):
            os.makedirs(self.root_folder)

        abspath = os.path.join(self.root_folder, file_name)

        try:
            open(abspath,"a").write(text+"\n")
            success = True
        except IOError as e:
            logging.error(
                "Error writing to %s (%s)" % (
                        abspath,
                        str(e)
                    )
                )
        except FileNotFoundError as e:
            logging.error(
                "Error writing to %s (%s)" % (
                        abspath,
                        str(e)
                    )
                )
        except PermissionError as e:
            logging.error(
                "Error writing to %s (%s)" % (
                        abspath,
                        str(e)
                    )
                )

        return success

    @staticmethod
    def write_to_file(file_name, text):
        success = False

        abspath = os.path.expanduser(file_name)

        try:
            with open(abspath, "a") as f:
                f.write(text + "\n")
                success = True
        except (IOError, FileNotFoundError, PermissionError) as e:
            raise Exception(
                f"Error writing to {abspath} ({str(e)})"
            )

        return success



import xml.etree.ElementTree as ET
from .templates import NAMESPACES
class ADWSError(Exception):
    """Custom exception for ADWS SOAP Faults."""
    def __init__(self, fault_data):
        self.raw_fault = fault_data if isinstance(fault_data, str) else fault_data.get("RawXML", "")
        self.detail_error = "Unknown ADWS Error"
        self.message = None
        self.errorcode = None
        self.fault_data = fault_data

        try:
            if isinstance(fault_data, str):
                # Handle raw XML string
                if fault_data.startswith("Sorting or Selection Property is invalid."):
                    xml_start = fault_data.find("<")
                    if xml_start != -1:
                        fault_data = fault_data[xml_start:]

                for prefix, uri in NAMESPACES.items():
                    ET.register_namespace(prefix, uri)

                root = ET.fromstring(fault_data)

                detail_element = root.find(".//soapenv:Detail", NAMESPACES) or root.find(".//s:Detail", NAMESPACES)
                if detail_element is not None:
                    dir_error = detail_element.find(".//ad:DirectoryError", NAMESPACES)
                    if dir_error is not None:
                        msg_elem = dir_error.find(".//ad:Message", NAMESPACES)
                        code_elem = dir_error.find(".//ad:ErrorCode", NAMESPACES)
                        if msg_elem is not None and msg_elem.text:
                            self.message = msg_elem.text.strip()
                        if code_elem is not None and code_elem.text:
                            try:
                                self.errorcode = int(code_elem.text.strip())
                            except Exception:
                                self.errorcode = code_elem.text.strip()
                        if self.message:
                            self.detail_error = self.message
                            super().__init__(self.detail_error)
                            return
                enum_fault = detail_element.find(".//ad:EnumerateFault", NAMESPACES)
                if enum_fault is not None:
                    error_element = enum_fault.find(".//ad:Error", NAMESPACES)
                    if error_element is not None and error_element.text:
                        self.detail_error = error_element.text.strip()
                        super().__init__(self.detail_error)
                        return

                reason_element = root.find(".//soapenv:Text", NAMESPACES) or root.find(".//s:Text", NAMESPACES)
                if reason_element is not None and reason_element.text:
                    self.detail_error = reason_element.text.strip()

            elif isinstance(fault_data, dict):
                # Handle dictionary data
                self.message = fault_data.get("Error")
                self.errorcode = fault_data.get("FaultCode")
                self.detail_error = fault_data.get("Error", "Unknown ADWS Error")

                if "ErrorDetail" in fault_data:
                    error_detail = fault_data["ErrorDetail"]
                    if "FaultDetail" in error_detail:
                        fault_detail = error_detail["FaultDetail"]
                        if "DirectoryError" in fault_detail:
                            dir_error = fault_detail["DirectoryError"]
                            self.message = dir_error.get("Message")
                            self.errorcode = dir_error.get("ErrorCode")
                            extended_error = dir_error.get("ExtendedErrorMessage")
                            if extended_error:
                                self.detail_error = extended_error
                            elif self.message:
                                self.detail_error = self.message
                        elif "ArgumentError" in fault_detail:
                            arg_error = fault_detail["ArgumentError"]
                            self.message = arg_error.get("Message")
                            # ArgumentError might not have a standard ErrorCode,
                            # but we can capture ShortMessage if available
                            self.errorcode = arg_error.get("ShortMessage")
                            if self.message:
                                self.detail_error = self.message
                        elif "ExtendedErrorMessage" in fault_detail:
                            self.detail_error = fault_detail["ExtendedErrorMessage"]
                        elif "ExtendedErrorDescription" in fault_detail:
                            self.detail_error = fault_detail["ExtendedErrorDescription"]
                    elif "Message" in error_detail:
                        self.message = error_detail.get("Message")
                        self.detail_error = self.message
            elif isinstance(fault_data, list):
                # Handle list data
                self.detail_error = "Multiple ADWS Errors"
                self.message = "Multiple ADWS Errors"
                self.errorcode = "Multiple"

        except ET.ParseError:
            self.detail_error = self.raw_fault.split("<")[0].strip() if "<" in self.raw_fault else self.raw_fault
        except Exception:
            pass

        super().__init__(self.detail_error)

    def __str__(self):
        return self.detail_error


#!/usr/bin/env python3
import logging
from impacket.dcerpc.v5 import transport, samr

class SamrObject:
	KNOWN_PROTOCOLS = {
        139: {'bindstr': r'ncacn_np:%s[\pipe\samr]', 'set_host': True},
        445: {'bindstr': r'ncacn_np:%s[\pipe\samr]', 'set_host': True},
    }

	def __init__(self, connection, port=445):
		self.connection = connection
		self.port = port
		self.pipetriggered = False

	def connect(self, target):
		stringBinding = self.KNOWN_PROTOCOLS[self.port]['bindstr'] % target
		dce = self.connection.connectRPCTransport(
			host=target,
			stringBindings=stringBinding,
			interface_uuid=samr.MSRPC_UUID_SAMR,
			raise_exceptions=True
		)
		return dce

	def open_handle(self, dce, builtin=False):
		index = 1 if builtin else 0
		server_handle = samr.hSamrConnect(dce)['ServerHandle']
		domain_name = samr.hSamrEnumerateDomainsInSamServer(dce, server_handle)['Buffer']['Buffer'][index]['Name']
		domain_id = samr.hSamrLookupDomainInSamServer(dce, server_handle, domain_name)['DomainId']
		domain_handle = samr.hSamrOpenDomain(dce, server_handle, domainId=domain_id)['DomainHandle']
		return domain_handle

	def get_object_rid(self, dce, domain_handle, object_name):
		object_id = None
		try:
			response = samr.hSamrLookupNamesInDomain(dce, domain_handle, (object_name,))
			object_id = response['RelativeIds']['Element'][0]['Data']
		except samr.DCERPCSessionError as e:
			if str(e).find('STATUS_MORE_ENTRIES') < 0:
				logging.error("[SAMR] No object found for {}".format(object_name))
		return object_id

	def get_user_handle(self, dce, domain_handle, user_rid):
		response = samr.hSamrOpenUser(dce, domain_handle, samr.MAXIMUM_ALLOWED, user_rid)
		return response['UserHandle']

	def close_handle(self, dce, handle):
		samr.hSamrCloseHandle(dce, handle)

	def get_all_local_users(self, dce, domain_handle):
		try:
			response = samr.hSamrEnumerateUsersInDomain(dce, domain_handle, samr.USER_NORMAL_ACCOUNT)
			for item in response['Buffer']['Buffer']:
				yield item
		except samr.DCERPCSessionError as e:
			if str(e).find('STATUS_MORE_ENTRIES') < 0:
				raise
		finally:
			self.close_handle(dce, domain_handle)

	def get_local_group(self, dce, domain_handle, gid):
		group_handle = samr.hSamrOpenGroup(dce, domain_handle, groupId=gid)['GroupHandle']
		group_sid = samr.hSamrRidToSid(dce, group_handle, gid)['Sid']
		si = samr.PSAMPR_SID_INFORMATION()
		si['SidPointer'] = group_sid
		self.close_handle(dce, group_handle)
		return si

	def get_local_user(self, dce, domain_handle, user_rid):
		try:
			user_handle = self.get_user_handle(dce, domain_handle, user_rid)
			response = samr.hSamrQueryInformationUser2(dce, user_handle, samr.USER_INFORMATION_CLASS.UserAllInformation)
			response = response['Buffer']['All']

			# Get groups that user is member of
			groups = samr.hSamrGetGroupsForUser(dce, user_handle)['Groups']['Groups']
			group_id_list = list(map(lambda g: g['RelativeId'], groups))
			
			sidArray = samr.SAMPR_PSID_ARRAY()
			for gid in group_id_list:
				si = self.get_local_group(dce, domain_handle, gid)
				sidArray['Sids'].append(si)

			global_lookup_ids = samr.hSamrLookupIdsInDomain(dce, domain_handle, group_id_list)
			response.fields['GlobalGroups'] = list(map(lambda a: a['Data'], global_lookup_ids['Names']['Element']))
			self.close_handle(dce, domain_handle)

			domain_handle = self.open_handle(dce, builtin=True)
			alias_membership = samr.hSamrGetAliasMembership(dce, domain_handle, sidArray)
			alias_id_list = list(map(lambda a: a['Data'], alias_membership['Membership']['Element']))

			local_lookup_ids = samr.hSamrLookupIdsInDomain(dce, domain_handle, alias_id_list)
			response.fields['LocalGroups'] = list(map(lambda a: a['Data'], local_lookup_ids['Names']['Element']))
			return response
		except samr.DCERPCSessionError as e:
			if str(e).find('STATUS_MORE_ENTRIES') < 0:
				raise
		finally:
			self.close_handle(dce, domain_handle)



		


#!/usr/bin/env python3

from ..utils.helpers import in_exception, LdapObfuscate
from ..utils.validators import (
	is_sid_attribute, is_dn_attribute, randomly_hex_encode_dn_string
)

def prepend_zeros_obfuscation(parsed_structure, transform_func, max_zeros=5):
	"""
	Prepend zeros obfuscation following Go RandPrependZerosFilterObf pattern.
	Adds leading zeros to numeric values and SID components.
	"""
	def obfuscate_prepend_zeros(attr, operator, value):
		if in_exception(attr):
			return attr, value
		
		if LdapObfuscate.is_number(value):
			return attr, LdapObfuscate.prepend_zeros_to_number(value, max_zeros)
		
		if is_sid_attribute(attr) and value.startswith("S-"):
			return attr, LdapObfuscate.prepend_zeros_to_sid(value, max_zeros)
		
		return attr, value

	transform_func(parsed_structure, obfuscate_prepend_zeros)

def hex_value_obfuscation(parsed_structure, transform_func, probability=0.5):
	"""
	Hex value obfuscation following Go RandHexValueFilterObf pattern.
	Applies hex encoding to DN string values.
	"""
	def obfuscate_hex_value(attr, operator, value):
		if in_exception(attr) or value == "*":
			return attr, value
		
		if is_dn_attribute(attr):
			return attr, randomly_hex_encode_dn_string(value, probability)
		
		return attr, value

	transform_func(parsed_structure, obfuscate_hex_value)

def spacing_obfuscation(parsed_structure, transform_func, max_spaces=3):
	"""
	Spacing obfuscation following Go RandSpacingFilterObf pattern.
	Adds context-aware spacing to attribute values.
	"""
	def obfuscate_spacing(attr, operator, value):
		if in_exception(attr):
			return attr, value
		
		if attr.lower() == "anr":
			return attr, LdapObfuscate.add_spacing_to_value(value, "anr", max_spaces)
		elif is_dn_attribute(attr):
			return attr, LdapObfuscate.add_spacing_to_value(value, "dn", max_spaces)
		elif is_sid_attribute(attr):
			return attr, LdapObfuscate.add_spacing_to_value(value, "sid", max_spaces)
		
		return attr, value

	transform_func(parsed_structure, obfuscate_spacing) 


#!/usr/bin/env python3
import json
import os
import hashlib
from datetime import datetime, timedelta
import logging
from ldap3.utils.ciDict import CaseInsensitiveDict
import base64
import tempfile

class Storage:
    def __init__(self):
        try:
            home_path = os.path.expanduser('~')
            if os.path.exists(home_path) and os.access(home_path, os.W_OK):
                self.root_folder = os.path.join(home_path, ".powerview")
                self.storage_folder = os.path.join(self.root_folder, "storage")
            else:
                self.root_folder = os.path.join(tempfile.gettempdir(), "powerview")
                self.storage_folder = os.path.join(self.root_folder, "storage")
            
            self.cache_folder = "ldap_cache"
            self.cache_path = os.path.join(self.storage_folder, self.cache_folder)
            
            os.makedirs(self.cache_path, mode=0o700, exist_ok=True)
            os.makedirs(self.root_folder, mode=0o700, exist_ok=True)
            
            logging.debug(f"[Storage] Using cache directory: {self.cache_path}")
            
        except Exception as e:
            temp_dir = tempfile.mkdtemp(prefix="powerview_")
            self.root_folder = temp_dir
            self.storage_folder = temp_dir
            self.cache_path = os.path.join(temp_dir, "ldap_cache")
            os.makedirs(self.cache_path, mode=0o700, exist_ok=True)
            logging.warning(f"[Storage] Using temporary directory for storage: {self.cache_path}")
            logging.error(f"[Storage] Original error: {e}")

    def _generate_cache_key(self, search_base, search_filter, search_scope, attributes, host, raw=False):
        """Generate a unique cache key based on search parameters"""
        cache_string = f"{search_base.lower()}|{search_filter.lower()}|{search_scope.lower()}|{str(sorted(attributes) if attributes else 'None')}|{host.lower()}|{raw}"
        return hashlib.md5(cache_string.encode()).hexdigest()

    def _serialize_complex_types(self, obj):
        """Serialize complex types like datetime, bytes, timedelta, and others."""
        if isinstance(obj, datetime):
            return {'__datetime__': obj.isoformat()}
        elif isinstance(obj, timedelta):
            return {'__timedelta__': obj.total_seconds()}
        elif isinstance(obj, bytes):
            return {'__bytes__': base64.b64encode(obj).decode('ascii')}
        elif isinstance(obj, dict):
            return {key: self._serialize_complex_types(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [self._serialize_complex_types(item) for item in obj]
        elif isinstance(obj, CaseInsensitiveDict):
            return {key: self._serialize_complex_types(value) for key, value in obj.items()}
        elif isinstance(obj, (tuple, set)):
            return list(obj)
        return obj

    def _deserialize_complex_types(self, obj):
        """Deserialize complex types like datetime, bytes, timedelta, and others."""
        if isinstance(obj, dict):
            if '__datetime__' in obj:
                return datetime.fromisoformat(obj['__datetime__'])
            elif '__timedelta__' in obj:
                return timedelta(seconds=obj['__timedelta__'])
            elif '__bytes__' in obj:
                return base64.b64decode(obj['__bytes__'].encode('ascii'))
            else:
                return {key: self._deserialize_complex_types(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [self._deserialize_complex_types(item) for item in obj]
        return obj

    def cache_results(self, search_base, search_filter, search_scope, attributes, host, results, raw=False):
        """Cache LDAP query results"""
        cache_key = self._generate_cache_key(search_base, search_filter, search_scope, attributes, host, raw)
        cache_file = os.path.join(self.cache_path, f"{cache_key}.json")

        try:
            serialized_results = self._serialize_complex_types(results)
            cache_data = {
                'timestamp': datetime.now().isoformat(),
                'raw': raw,
                'search_base': search_base,
                'search_filter': search_filter,
                'search_scope': search_scope,
                'attributes': attributes,
                'host': host,
                'results': serialized_results
            }

            with open(cache_file, 'w') as f:
                json.dump(cache_data, f, indent=4)
        except Exception as e:
            logging.error(f"Error caching results: {e}")

    def get_cached_results(self, search_base, search_filter, search_scope, attributes, host, cache_ttl=1800, raw=False):
        """Retrieve cached LDAP query results if they exist and are not expired"""
        cache_key = self._generate_cache_key(search_base, search_filter, search_scope, attributes, host, raw)
        cache_file = os.path.join(self.cache_path, f"{cache_key}.json")

        try:
            if os.path.exists(cache_file):
                with open(cache_file, 'r') as f:
                    cached_data = json.load(f)
                
                cache_time = datetime.fromisoformat(cached_data['timestamp'])
                if datetime.now() - cache_time < timedelta(seconds=cache_ttl):
                    return self._deserialize_complex_types(cached_data['results'])
                else:
                    try:
                        os.remove(cache_file)
                        logging.debug(f"[Storage] Deleted expired cache file: {cache_key}")
                    except Exception as e:
                        logging.error(f"[Storage] Error deleting expired cache file: {e}")
        except Exception as e:
            logging.error(f"[Storage] Error reading cache: {e}")
        
        return None

    def clear_cache(self) -> bool:
        """Clear all cached LDAP results"""
        try:
            for file in os.listdir(self.cache_path):
                if file.endswith('.json'):
                    os.remove(os.path.join(self.cache_path, file))
            return True
        except Exception as e:
            logging.error(f"Error clearing cache: {e}")
            return False


#!/usr/bin/env python3

import logging

from impacket.ldap import ldaptypes
from impacket.uuid import bin_to_string
from powerview.modules.ldapattack import ACE_FLAGS, OBJECT_ACE_FLAGS, SIMPLE_PERMISSIONS, ACCESS_MASK

class AccessControl:
	@staticmethod
	def get_default_sd():
		return AccessControl.create_empty_sd()

	@staticmethod
	def create_empty_sd(
		owner_sid: str='S-1-5-32-544', 
		control: int=32772,
		acl_revision: int=4,
		sbz1: int=0,
		sbz2: int=0
	):
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		sd['Revision'] = b'\x01'
		sd['Sbz1'] = b'\x00'
		sd['Control'] = control
		sd['OwnerSid'] = ldaptypes.LDAP_SID()
		sd['OwnerSid'].fromCanonical(owner_sid)
		sd['GroupSid'] = b''
		sd['Sacl'] = b''
		acl = ldaptypes.ACL()
		acl['AclRevision'] = acl_revision
		acl['Sbz1'] = sbz1
		acl['Sbz2'] = sbz2
		acl.aces = []
		sd['Dacl'] = acl
		return sd

	@staticmethod
	def create_ace(
		sid: str, 
		mask: int=983551, 
		ace_type: int=ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE, 
		ace_flags: int=0x00
	):
		nace = ldaptypes.ACE()
		nace['AceType'] = ace_type
		nace['AceFlags'] = ace_flags
		acedata = ldaptypes.ACCESS_ALLOWED_ACE()
		acedata['Mask'] = ldaptypes.ACCESS_MASK()
		acedata['Mask']['Mask'] = mask
		acedata['Sid'] = ldaptypes.LDAP_SID()
		acedata['Sid'].fromCanonical(sid)
		nace['Ace'] = acedata
		return nace

	@staticmethod
	def parse_perms(fsr):
		_perms = []
		for PERM in SIMPLE_PERMISSIONS:
			if (fsr & PERM.value) == PERM.value:
				_perms.append(PERM.name)
				fsr = fsr & (not PERM.value)
		for PERM in ACCESS_MASK:
			if fsr & PERM.value:
				_perms.append(PERM.name)
		return _perms

	@staticmethod
	def get_user_sid(secDesc):
		sids = []
		if not secDesc:
			return

		for dacl in ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDesc)['Dacl']['Data']:
			sids.append(dacl['Ace']['Sid'].formatCanonical())
		return sids
		
	@staticmethod
	def parse_sd(secDesc):
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		if isinstance(secDesc, list):
			secDesc = b''.join(secDesc)
		sd.fromString(secDesc)

		security_info = {}
		if sd['OwnerSid'] is not None:
			security_info['OwnerSid'] = sd['OwnerSid'].formatCanonical()
		if sd['GroupSid'] is not None:
			security_info['GroupSid'] = sd['GroupSid'].formatCanonical()

		try:
			dacl_data = sd['Dacl']['Data']
			aces = []
			for ace_obj in dacl_data:
				ace_type_name = ace_obj['TypeName']
				
				if ace_type_name not in ["ACCESS_ALLOWED_ACE", "ACCESS_DENIED_ACE", 
											"ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_OBJECT_ACE",
											"SYSTEM_AUDIT_ACE", "SYSTEM_ALARM_ACE",
											"SYSTEM_AUDIT_OBJECT_ACE", "SYSTEM_ALARM_OBJECT_ACE"]:
					logging.debug(f"[SMBClient: get_file_info] Skipping unhandled ACE type: {ace_type_name}")
					continue

				trustee_sid_str = ace_obj['Ace']['Sid'].formatCanonical()
				
				ace_flags_int = ace_obj['AceFlags']
				parsed_ace_flags_list = [FLAG.name for FLAG in ACE_FLAGS if ace_flags_int & FLAG.value]

				access_mask_int = ace_obj['Ace']['Mask']['Mask']
				parsed_permissions_list = AccessControl.parse_perms(access_mask_int)
				
				if not parsed_permissions_list and access_mask_int != 0: # If no known flags matched but mask is not zero
					parsed_permissions_list.append(f"UNKNOWN_MASK_0x{access_mask_int:08X}")

				# Initialize object-specific fields
				parsed_object_ace_specific_flags_list = None
				obj_type_guid_str = None
				inh_obj_type_guid_str = None

				if ace_type_name in ["ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_OBJECT_ACE", "SYSTEM_AUDIT_OBJECT_ACE", "SYSTEM_ALARM_OBJECT_ACE"]:
					object_ace_specific_flags_int = ace_obj['Ace']['Flags']
					parsed_object_ace_specific_flags_list = [FLAG.name for FLAG in OBJECT_ACE_FLAGS if object_ace_specific_flags_int & FLAG.value]
					
					if ace_obj['Ace']['ObjectTypeLen'] != 0:
						obj_type_guid_str = bin_to_string(ace_obj['Ace']['ObjectType']).lower()
					
					if ace_obj['Ace']['InheritedObjectTypeLen'] != 0:
						inh_obj_type_guid_str = bin_to_string(ace_obj['Ace']['InheritedObjectType']).lower()

				ace_info_entry = {
					'type': ace_type_name,
					'trustee': trustee_sid_str,
					'ace_flags': parsed_ace_flags_list,
					'access_mask_raw': access_mask_int,
					'permissions': parsed_permissions_list,
					'object_ace_specific_flags': parsed_object_ace_specific_flags_list,
					'object_type_guid': obj_type_guid_str,
					'inherited_object_type_guid': inh_obj_type_guid_str
				}
				aces.append(ace_info_entry)

			security_info['Dacl'] = aces
			return security_info
		except Exception as e:
			logging.error(f"[SMBClient: get_file_info] Error parsing security descriptor: {e}")
			import traceback
			logging.debug(f"[SMBClient: get_file_info] Traceback: {traceback.format_exc()}")
			if 'security_info' not in locals(): security_info = {}
			if 'Dacl' not in security_info : security_info['Dacl'] = []
			return security_info

	@staticmethod
	def sort_dacl_aces(sd):
		"""
		Sort ACEs in the DACL according to Windows canonical order:
		1. Explicit DENY ACEs
		2. Explicit ALLOW ACEs  
		3. Inherited DENY ACEs
		4. Inherited ALLOW ACEs
		"""
		if not sd['Dacl'] or not hasattr(sd['Dacl'], 'aces'):
			return sd
			
		aces = sd['Dacl'].aces
		if not aces:
			return sd
			
		def ace_sort_key(ace):
			ace_type = ace['AceType']
			ace_flags = ace['AceFlags']
			
			is_inherited = bool(ace_flags & 0x10)
			
			is_deny = ace_type in [
				ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE,
				ldaptypes.ACCESS_DENIED_OBJECT_ACE.ACE_TYPE
			]
			
			if not is_inherited and is_deny:
				return (0, ace_type)
			elif not is_inherited and not is_deny:
				return (1, ace_type)
			elif is_inherited and is_deny:
				return (2, ace_type)
			else:
				return (3, ace_type)
		
		sorted_aces = sorted(aces, key=ace_sort_key)
		sd['Dacl'].aces = sorted_aces
		
		return sd

	@staticmethod
	def add_allow_ace(
		secDesc: bytes,
		sid: str,
		mask: int=SIMPLE_PERMISSIONS.FullControl.value, # Full Control
		inherit: bool=True
	):
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		if isinstance(secDesc, list):
			secDesc = b''.join(secDesc)
		sd.fromString(secDesc)
		ace_flags = 0x00
		if inherit:
			ace_flags = ACE_FLAGS.CONTAINER_INHERIT_ACE.value + ACE_FLAGS.OBJECT_INHERIT_ACE.value
		sd['Dacl'].aces.append(
			AccessControl.create_ace(
				sid,
				mask,
				ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE,
				ace_flags
			)
		)
		sd = AccessControl.sort_dacl_aces(sd)
		return sd.getData()

	@staticmethod
	def add_deny_ace(
		secDesc: bytes,
		sid: str,
		mask: int=SIMPLE_PERMISSIONS.FullControl.value,
		inherit: bool=True
	):
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		if isinstance(secDesc, list):
			secDesc = b''.join(secDesc)
		sd.fromString(secDesc)
		ace_flags = 0x00
		if inherit:
			ace_flags = ACE_FLAGS.CONTAINER_INHERIT_ACE.value + ACE_FLAGS.OBJECT_INHERIT_ACE.value
		sd['Dacl'].aces.append(
			AccessControl.create_ace(
				sid,
				mask,
				ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE,
				ace_flags
			)
		)
		sd = AccessControl.sort_dacl_aces(sd)
		return sd.getData()

	@staticmethod
	def remove_ace(
		secDesc: bytes,
		sid: str,
		mask: int=None,
		ace_type: int=None
	):
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		if isinstance(secDesc, list):
			secDesc = b''.join(secDesc)
		sd.fromString(secDesc)
		
		aces_to_remove = []
		for ace in sd['Dacl'].aces:
			ace_sid = ace['Ace']['Sid'].formatCanonical()
			ace_mask = ace['Ace']['Mask']['Mask']
			ace_type_val = ace['AceType']
			
			if ace_sid == sid:
				if mask is None or ace_mask == mask:
					if ace_type is None or ace_type_val == ace_type:
						aces_to_remove.append(ace)
		
		for ace in aces_to_remove:
			sd['Dacl'].aces.remove(ace)
		
		sd = AccessControl.sort_dacl_aces(sd)
		
		return sd.getData(), len(aces_to_remove)

	@staticmethod
	def fix_dacl_ordering(secDesc: bytes):
		"""
		Fix the ordering of ACEs in an existing DACL to match Windows canonical order.
		This is useful for cleaning up DACLs that may have been created with incorrect ordering.
		"""
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
		if isinstance(secDesc, list):
			secDesc = b''.join(secDesc)
		sd.fromString(secDesc)
		
		sd = AccessControl.sort_dacl_aces(sd)
		
		return sd.getData()


# https://github.com/Wh04m1001/DFSCoerce/blob/main/dfscoerce.py
from impacket import system_errors
from impacket.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT
from impacket.dcerpc.v5.dtypes import UUID, ULONG, WSTR, DWORD
from impacket.dcerpc.v5.rpcrt import DCERPCException
from impacket.uuid import uuidtup_to_bin

MSRPC_UUID_DFSNM = uuidtup_to_bin(('4FC742E0-4A10-11CF-8273-00AA004AE673', '3.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__(self):
		key = self.error_code
		if key in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key][1]
			return 'DFSNM SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		else:
			return 'DFSNM SessionError: unknown error code: 0x%x' % self.error_code


class NetrDfsRemoveStdRoot(NDRCALL):
	opnum = 13
	structure = (
		('ServerName', WSTR),
		('RootShare', WSTR),
		('ApiFlags', DWORD),
	)


class NetrDfsRemoveStdRootResponse(NDRCALL):
	structure = (
		('ErrorCode', ULONG),
	)
class NetrDfsAddRoot(NDRCALL):
	opnum = 12
	structure = (
		 ('ServerName',WSTR),
		 ('RootShare',WSTR),
		 ('Comment',WSTR),
		 ('ApiFlags',DWORD),
	 )
class NetrDfsAddRootResponse(NDRCALL):
	 structure = (
		 ('ErrorCode', ULONG),
	 )


#!/usr/bin/env python3
from powerview.utils.colors import bcolors, Gradient

def get_prompt(powerview, current_target_domain=None, using_cache=False, args=None):
	"""
	Creates a visually enhanced prompt for the PowerView shell.
	
	Args:
		powerview: PowerView instance to extract connection and status information from
		current_target_domain: Optional target domain for cross-domain operations
		using_cache: Indicates if the last results came from cache
		args: Command line arguments for additional context
	Returns:
		A formatted string for the shell prompt
	"""
	try:
		init_proto = powerview.conn.get_proto()
		server_dns = powerview.get_server_dns()
		nameserver = powerview.conn.get_nameserver()
		obfuscate = getattr(args, 'obfuscate', False)
		
		is_admin = False
		if args and not getattr(args, 'no_admin_check', False):
			is_admin = powerview.get_admin_status()
		
		cur_user = powerview.conn.who_am_i()
		if is_admin:
			cur_user = f"{bcolors.WARNING}{cur_user}{bcolors.ENDC}"
		
		mcp_running = False
		web_running = False
		if args:
			if getattr(args, 'mcp', False) and hasattr(powerview, 'mcp_server'):
				mcp_running = powerview.mcp_server.get_status()
			if getattr(args, 'web', False) and hasattr(powerview, 'api_server'):
				web_running = powerview.api_server.get_status()
		
		channel_binding_active = getattr(powerview.conn, 'use_channel_binding', False)
		ldap_signing_active = getattr(powerview.conn, 'use_sign_and_seal', False)
		
	except Exception as e:
		return f"{bcolors.OKGREEN}PV ❯{bcolors.ENDC} "
	
	domain_indicator = ""
	if current_target_domain:
		domain_indicator = f" {bcolors.BOLD}{bcolors.FAIL}[→ {current_target_domain}]{bcolors.ENDC}"
	
	cache_indicator = ""
	if using_cache:
		cache_indicator = f" {bcolors.WARNING}[CACHED]{bcolors.ENDC}"
	
	mcp_indicator = ""
	if mcp_running:
		mcp_text = "[MCP]"
		gradient_colors = Gradient.generate_gradient_colors([138, 43, 226], [0, 191, 255], len(mcp_text))
		colored_text = ""
		for i, char in enumerate(mcp_text):
			r, g, b = gradient_colors[i]
			colored_text += f"\033[38;2;{r};{g};{b}m{char}\033[0m"
		mcp_indicator = f" {bcolors.BOLD}{colored_text}{bcolors.ENDC}"

	web_indicator = ""
	if web_running:
		web_indicator = f" {bcolors.OKBLUE}[WEB]{bcolors.ENDC}"
	
	security_indicators = ""
	if channel_binding_active:
		security_indicators += "📦 "
	if ldap_signing_active:
		security_indicators += "🔒 "
	if obfuscate:
		security_indicators += "😈 "
	
	prompt = (f'{bcolors.OKBLUE}╭─{bcolors.ENDC}'
			  f'{security_indicators}'
			  f'{bcolors.WARNING}{bcolors.BOLD}{init_proto}{bcolors.ENDC}'
			  f'{bcolors.OKBLUE}─[{bcolors.ENDC}{bcolors.OKCYAN}{server_dns}{bcolors.ENDC}{bcolors.OKBLUE}]{bcolors.ENDC}'
			  f'{bcolors.OKBLUE}─[{bcolors.ENDC}{cur_user}{bcolors.OKBLUE}]{bcolors.ENDC}'
			  f'{bcolors.OKBLUE}-[NS:{nameserver if nameserver else "<auto>"}]{bcolors.ENDC}'
			  f'{mcp_indicator}'
			  f'{web_indicator}'
			  f'{domain_indicator}'
			  f'{cache_indicator}'
			  f'\n{bcolors.OKBLUE}╰─{bcolors.BOLD}PV{bcolors.ENDC} {bcolors.OKGREEN}❯{bcolors.ENDC} ')
	
	return prompt


# This code is adapted from https://github.dev/xforcered/SoaPy/
import socket
from enum import IntEnum
from typing import Type

import impacket.structure

from .encoder import Encoder
from .nns import NNS
from .core.server import Server

"""
[MC-NMF]: .NET Message Framing Protocol.  This is the initiator (client) implementation.
Only Duplex mode is implmented.

"""


class RecordType(IntEnum):
	"""
	Protocol for record exchange

	[MC-NMF]: 2.2.1
	"""

	VERSION = 0x0
	MODE = 0x1
	VIA = 0x2
	KNOWN_ENCODING = 0x3
	EXTENSIBLE_ENCODING = 0x4
	UNSIZED_ENVELOPE = 0x5
	SIZED_ENVELOPE = 0x6
	END = 0x7
	FAULT = 0x8
	UPGRADE_REQUEST = 0x9
	UPGRADE_RESPONSE = 0xA
	PREAMBLE_ACK = 0xB
	PREAMBLE_END = 0xC


class Mode(IntEnum):
	"""Communication mode

	[MC-NMF]: 2.2.3.2
	"""

	SINGELTON_UNSIZED = 0x1
	DUPLEX = 0x2
	SIMPLEX = 0x3
	SINGLETON_SIZED = 0x4


class KnownEncoding(IntEnum):
	"""Encoding Envelope Records

	[MC-NMF]: 2.2.3.4.1
	"""

	# soap 1.1
	SOAP1_1_UTF8 = 0x0  # [RFC2279]
	SOAP1_1_UTF16 = 0x1  # [RFC2781]
	SOAP1_1_UNICODE_LE = 0x2

	# soap 1.2
	SOAP1_2_UTF8 = 0x3
	SOAP1_2_UTF16 = 0x4
	SOAP1_2_UNICODE = 0x5
	SOAP1_2_MTOM = 0x6  # [SOAP-MTOM]
	SOAP1_2_BINARY = 0x7  # [MC-NBFS]
	SOAP1_2_BINARY_INBAND_DICT = 0x8  # [MC-NBFSE]


class NMFServerFault(Exception): ...


class NMFUnknownRecord:
	def __init__(self, data=""):
		raise NMFServerFault(f"NMFUnknownRecord type {data}")


class NMFRecord(impacket.structure.Structure):
	structure: tuple[tuple[str, str] | tuple[str, str, object], ...]

	def send(self, sock: socket.socket | NNS):
		sock.sendall(self.getData())

	@staticmethod
	def encode_size(size: int) -> bytes:
		"""NMF variable size encoding for records.

		[MC-NMF]: 2.2.2

		Args:
			size (int): size of the record payload

		Returns:
			bytes: packed and encoded size as bytes
		"""
		MAXMBI = 0x7F
		if size < 0:
			raise ValueError("Signed numbers are not supported")

		if size <= MAXMBI:
			return bytes([size])

		result = []
		for _ in range(5):
			byte = size & MAXMBI
			size >>= 7
			if size != 0:
				byte |= MAXMBI + 1
			result.append(byte)

			if size == 0:
				break

		return bytes(result)

	@staticmethod
	def decode_size(encoded_data: bytes) -> tuple[int, int, bytes]:
		"""NMF decode size of record payload.

		Returns the size of the payload, and also the number of bytes the size
		value takes.  The size field is variable length.

		To use this method, first slice off the record type so that the first
		field in the encoded_data is the size feild

		The size feild can be max 5 bytes long.

		[MC-NMF]: 2.2.2

		Args:
			encoded_data (bytes): payload with size included

		Returns:
			tuple[int, int, bytes]: size of the payload, the number of bytes in the size field, the payload
		"""

		size = 0
		shift = 0
		len_length = 0

		max_len = min(len(encoded_data), 5)

		for i in range(max_len):
			byte = encoded_data[i]

			len_length += 1
			size |= (byte & 0x7F) << shift
			if not (byte & 0x80):
				return size, len_length, encoded_data[len_length:]
			shift += 7

			if size > 0xFFFFFFFF:
				raise ValueError("Size too big")

		return size, 1, encoded_data[1:]


class NMFVersion(NMFRecord):
	structure = (
		("record_type", ">B"),
		("major_version", ">B"),
		("minor_version", ">B"),
	)

	def __init__(
		self, major_version: int = 0, minor_version: int = 0, data: None | bytes = None
	):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.VERSION
			self["major_version"] = major_version
			self["minor_version"] = minor_version


class NMFMode(NMFRecord):
	structure = (
		("record_type", ">B"),
		("mode", ">B"),
	)

	def __init__(self, mode: int = 0, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.MODE
			self["mode"] = mode


class NMFVia(NMFRecord):
	structure = (
		("record_type", ">B"),
		("via_len", ":"),  # warning, this is variable length field
		("via", ":"),
	)

	def __init__(self, via: str = "", data: None | bytes = None):
		"""This record has variable length field encoding on the
		via_len field.  Reading 'via_len' will not return a correct
		value.

		Args:
			via (str): via to send
			data (bytes): a packed full length record
		"""
		impacket.structure.Structure.__init__(self)
		if data:
			self["record_type"] = data[0]
			_, _, payload = self.decode_size(data[1:])
			self["via"] = payload.decode("utf-8")
		else:
			self["record_type"] = RecordType.VIA
			self["via_len"] = self.encode_size(len(via.encode("utf-8")))
			self["via"] = via.encode("utf-8")


class NMFKnownEncoding(NMFRecord):
	structure = (
		("record_type", ">B"),
		("encoding", ">B"),
	)

	def __init__(self, encoding: int = 0, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.KNOWN_ENCODING
			self["encoding"] = encoding


class NMFExtensibleEncoding(NMFRecord): ...


class NMFSizedEnvelope(NMFRecord):
	structure = (
		("record_type", ">B"),
		("size", ":"),
		("payload", ":"),
	)

	def __init__(self, payload: bytes = bytes(), data: None | bytes = None):
		"""This record has variable length field encoding on the
		'size' field.  Reading 'size' will not return a correct
		value.

		Args:
			payload (bytes): payload to send
			data (bytes): a packed sized envelope record
		"""
		impacket.structure.Structure.__init__(self, data=data)
		if data:
			self["record_type"] = data[0]
			_, _, env_payload = self.decode_size(data[1:])
			self["payload"] = env_payload
		else:
			self["record_type"] = RecordType.SIZED_ENVELOPE
			self["size"] = self.encode_size(len(payload))
			self["payload"] = payload


class NMFEnd(NMFRecord):
	structure = (("record_type", ">B"),)

	def __init__(self, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.END


class NMFUnsizedEnvelope(NMFRecord): ...


class NMFFault(NMFRecord):
	structure = (
		("record_type", ">B"),
		("size", ":"),
		("fault", ":"),
	)

	def __init__(self, fault: str = "", data: None | bytes = None):
		"""This record has variable length field encoding on the
		'size' field.  Reading 'size' will not return a correct
		value.

		Args:
			fault (str): fault msg to send
			data (bytes): a packed full length record
		"""
		impacket.structure.Structure.__init__(self, data=data)
		if data:
			self["record_type"] = data[0]
			_, _, payload = self.decode_size(data[1:])
			self["fault"] = payload.decode("utf-8")

		else:
			self["record_type"] = RecordType.FAULT
			self["size"] = self.encode_size(len(fault.encode("utf-8")))
			self["fault"] = fault.encode("utf-8")


class NMFUpgradeRequest(NMFRecord):
	structure = (
		("record_type", ">B"),
		("proto_len", ":"),
		("proto", ":"),
	)

	def __init__(self, proto: str = "application/negotiate", data: None | bytes = None):
		"""This record has variable length field encoding on the
		'proto_len' field.  Reading 'proto_len' will not return a correct
		value.

		Args:
			proto (str): proto to send
			data (bytes): a packed full length record
		"""

		impacket.structure.Structure.__init__(self, data=data)

		if data:
			self["record_type"] = data[0]
			_, _, payload = self.decode_size(data[1:])
			self["proto"] = payload.decode("utf-8")
		else:
			self["record_type"] = RecordType.UPGRADE_REQUEST
			self["proto_len"] = self.encode_size(len(proto.encode("utf-8")))
			self["proto"] = proto.encode("utf-8")


class NMFUpgradeResponse(NMFRecord):
	structure = (("record_type", ">B"),)

	def __init__(self, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.UPGRADE_RESPONSE


class NMFPreambleEnd(NMFRecord):
	structure = (("record_type", ">B"),)

	def __init__(self, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.PREAMBLE_END


class NMFPreambleAck(NMFRecord):
	structure = (("record_type", ">B"),)

	def __init__(self, data: None | bytes = None):
		impacket.structure.Structure.__init__(self, data=data)
		if not data:
			self["record_type"] = RecordType.PREAMBLE_ACK


class NMFPreamble(NMFRecord):
	structure = (
		("version", ":"),
		("mode", ":"),
		("via", ":"),
		("encoding", ":"),
	)

	def __init__(
		self,
		version: tuple[int, int] = (1, 1),
		mode: int = 0,
		via: str = "",
		encoding: int = 0,
	):
		impacket.structure.Structure.__init__(self)
		self["version"] = NMFVersion(*version).getData()
		self["mode"] = NMFMode(mode).getData()
		self["via"] = NMFVia(via).getData()
		self["encoding"] = NMFKnownEncoding(encoding).getData()


class NMFConnection:
	def __init__(
		self,
		nns: NNS,
		fqdn: str,
		port: int = 9389,
		mode: int = Mode.DUPLEX,
		encoding: int = KnownEncoding.SOAP1_2_BINARY_INBAND_DICT,
	):
		"""
		Args:
			nns (NNS): NNS Connection object
			fqdn (str): FQDN of endpoint we wish to talk to
		"""

		self._nns: NNS = nns
		self._sock: socket.socket = nns._sock

		# before being upgraded, the transport is raw socket
		self._transport: NNS | socket.socket = self._sock

		self._mode = mode
		self._encoding = encoding
		self._fqdn = fqdn
		self._port = port

		self._encoder = Encoder(self._encoding)

	def _throw_if_not(self, expected: Type[NMFRecord], got: NMFRecord):
		"""Allows for quick validation of expected responses.  If not expected
		checks for fault, and throws

		Args:
			expected (NMFRecord): the type you expected
			got (NMFRecord): what you have
		"""

		if not isinstance(got, expected):
			if isinstance(got, NMFFault):
				raise ConnectionError(got["fault"])
			raise ConnectionError(
				f"Unexpected server response.  Expected  {expected['record_type']}, got {got['record_type']}"
			)

	def _upgrade(self):
		"""Upgrade if using NNS for transport,
		otherwise do nothing.  This allows for support
		of unauthenticated transport for things like MEX
		"""

		if not isinstance(self._nns, NNS):
			return

		NMFUpgradeRequest().send(self._transport)

		# wait for ack
		self._throw_if_not(NMFUpgradeResponse, self._recv())

		# nns auth
		self._nns.auth_ntlm()

		# switch to upgraded transport now
		self._transport = self._nns

	def connect(self, resource: str):
		"""Establish connection to server. Set up all the communication
		channels that are nessisary to begin data exchanges.

		Send NMF preamble
		Upgrade to NNS
		NNS Auth
		End NMF preamble

		Args:
			resource (str): Resource to request in the via

		Raises:
			NMFServerFault: Raises when the server ack upgrade request
			or the preamble end, indicating connection falure.
		"""

		# send the preamble
		NMFPreamble(
			version=(1, 0),
			mode=self._mode,
			via=f"net.tcp://{self._fqdn}:9389/ActiveDirectoryWebServices/{resource}",
			encoding=self._encoding,
		).send(self._transport)

		self._upgrade()

		# preamble end
		NMFPreambleEnd().send(self._transport)

		# wait for ack
		self._throw_if_not(NMFPreambleAck, self._recv())

	def _end_record(self) -> None:
		"""Send an end record"""
		NMFEnd().send(self._transport)

	def send(self, data: str):
		"""Send data to server in an envelope msg.  This assumes that
		the current mode is duplex or simplex.

		Args:
			data (str): data to send
		"""
		encoding_data: bytes = self._encoder.encode(data)
		NMFSizedEnvelope(payload=encoding_data).send(self._transport)

	def recv(self) -> str:
		"""Receive data from the transport mechanism.  Automatically
		decodes the data based on selected transport encoding
		type.

		Returns:
			(str): received and decoded data
		Raises:
			NMFServerFault: Raises if not data transport msg
		"""
		pkt = self._recv()

		self._throw_if_not(NMFSizedEnvelope, pkt)

		return self._encoder.decode(pkt["payload"])

	def _recv(self) -> NMFRecord:
		"""Read a packet from the network transport layer and
		return the object version of the packet

		Returns:
			NMFRecord: Correct NMF object
		"""

		jump_table = {
			0x0: NMFVersion,
			0x1: NMFMode,
			0x2: NMFVia,
			0x3: NMFKnownEncoding,
			0x4: NMFExtensibleEncoding,
			0x5: NMFUnsizedEnvelope,
			0x6: NMFSizedEnvelope,
			0x7: NMFEnd,
			0x8: NMFFault,
			0x9: NMFUpgradeRequest,
			0xA: NMFUpgradeResponse,
			0xB: NMFPreambleAck,
			0xC: NMFPreambleEnd,
		}

		data: bytes = self._transport.recv(4096)

		record_type: int = data[0]

		# simplifed object factory, takes type and returns NMFRecords
		return jump_table.get(record_type, NMFUnknownRecord)(data=data)



#!/usr/bin/env python3

from powerview.utils.helpers import IDict
from powerview.utils.constants import ATTRIBUTE_OID

WILDCARD = "*"
COMMA = ","
COLON = ":"
MAX_RAND = 10

ATTRIBUTE_OID = IDict(ATTRIBUTE_OID)

EXCEPTION_ATTRIBUTES = [
	"objectClass",
	"objectCategory",
	"userAccountControl",
	"msDS-AllowedToActOnBehalfOfOtherIdentity"
]

EXCEPTION_CHARS = [
	WILDCARD,
	COMMA,
	COLON
]

EXACT_MATCH_ATTRIBUTES = [
	'samaccounttype', 'admincount', 'useraccountcontrol', 
	'primarygroupid', 'objectclass', 'objectcategory',
	'instancetype', 'systemflags', 'options'
]

DEFAULT_ANR_ATTRIBUTES = [
	'samaccountname', 'displayname', 'name', 'cn', 'commonname',
	'givenname', 'surname', 'mail', 'userprincipalname'
]

SID_ATTRIBUTES = ['objectsid', 'sid', 'msds-allowedtoactonbehalfofotheridentity']

DN_ATTRIBUTES = ['distinguishedname', 'dn', 'manager', 'member', 'memberof']

BITWISE_ATTRIBUTES = ['useraccountcontrol', 'systemflags', 'instancetype', 'options']

WILDCARD_SAFE_ATTRIBUTES = ['name', 'cn', 'displayname', 'description', 'mail', 'userprincipalname'] 


from impacket.dcerpc.v5 import tsts as TSTS
from impacket.dcerpc.v5 import transport, lsat, lsad
from impacket.dcerpc.v5.dtypes import MAXIMUM_ALLOWED
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_GSS_NEGOTIATE, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException
import logging
import traceback

class TSHandler:
	def __init__(self, smb_connection, target_ip, doKerberos=False):
		self.__target_ip = target_ip
		self.__doKerberos = doKerberos
		self.__smbConnection = smb_connection

	def get_session_list(self):
		# Retreive session list
		with TSTS.TermSrvEnumeration(self.__smbConnection, self.__target_ip, self.__doKerberos) as lsm:
			handle = lsm.hRpcOpenEnum()
			rsessions = lsm.hRpcGetEnumResult(handle, Level=1)['ppSessionEnumResult']
			lsm.hRpcCloseEnum(handle)
			self.sessions = {}
			for i in rsessions:
				sess = i['SessionInfo']['SessionEnum_Level1']
				state = TSTS.enum2value(TSTS.WINSTATIONSTATECLASS, sess['State']).split('_')[-1]
				self.sessions[sess['SessionId']] = { 'state'        :state,
													'SessionName'   :sess['Name'],
													'RemoteIp'      :'',
													'ClientName'    :'',
													'Username'      :'',
													'Domain'        :'',
													'Resolution'    :'',
													'ClientTimeZone':''
												}

	def enumerate_sessions_config(self):
		# Get session config one by one
		if len(self.sessions):
			with TSTS.RCMPublic(self.__smbConnection, self.__target_ip, self.__doKerberos) as termsrv:
				for SessionId in self.sessions:
					resp = termsrv.hRpcGetClientData(SessionId)
					if resp is not None:
						self.sessions[SessionId]['RemoteIp']       = resp['ppBuff']['ClientAddress']
						self.sessions[SessionId]['ClientName']     = resp['ppBuff']['ClientName']
						if len(resp['ppBuff']['UserName']) and not len(self.sessions[SessionId]['Username']):
							self.sessions[SessionId]['Username']   = resp['ppBuff']['UserName']
						if len(resp['ppBuff']['Domain']) and not len(self.sessions[SessionId]['Domain']):
							self.sessions[SessionId]['Domain']     = resp['ppBuff']['Domain']
						self.sessions[SessionId]['Resolution']     = '{}x{}'.format(
																		resp['ppBuff']['HRes'],
																		resp['ppBuff']['VRes']
																	)
						self.sessions[SessionId]['ClientTimeZone'] = resp['ppBuff']['ClientTimeZone']['StandardName']

	def enumerate_sessions_info(self):
		# Get session info one by one
		if len(self.sessions):
			with TSTS.TermSrvSession(self.__smbConnection, self.__target_ip, self.__doKerberos) as TermSrvSession:
				for SessionId in self.sessions.keys():
					sessdata = TermSrvSession.hRpcGetSessionInformationEx(SessionId)
					sessflags = TSTS.enum2value(TSTS.SESSIONFLAGS, sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['SessionFlags'])
					self.sessions[SessionId]['flags']    = sessflags
					domain = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['DomainName']
					if not len(self.sessions[SessionId]['Domain']) and len(domain):
						self.sessions[SessionId]['Domain'] = domain
					username = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['UserName']
					if not len(self.sessions[SessionId]['Username']) and len(username):
						self.sessions[SessionId]['Username'] = username
					self.sessions[SessionId]['ConnectTime'] = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['ConnectTime']
					self.sessions[SessionId]['DisconnectTime'] = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['DisconnectTime']
					self.sessions[SessionId]['LogonTime'] = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['LogonTime']
					self.sessions[SessionId]['LastInputTime'] = sessdata['LSMSessionInfoExPtr']['LSM_SessionInfo_Level1']['LastInputTime']

	def do_qwinsta(self):
		desktop_states = {
			'WTS_SESSIONSTATE_UNKNOWN': '',
			'WTS_SESSIONSTATE_LOCK'   : 'Locked',
			'WTS_SESSIONSTATE_UNLOCK' : 'Unlocked',
		}
		self.get_session_list()
		if not len(self.sessions):
			return []
		self.enumerate_sessions_info()
		self.enumerate_sessions_config()
		
		result = []
		
		for i in self.sessions:
			connectTime = self.sessions[i]['ConnectTime']
			connectTime = connectTime.strftime(r'%Y/%m/%d %H:%M:%S') if connectTime.year > 1601 else 'None'

			disconnectTime = self.sessions[i]['DisconnectTime']
			disconnectTime = disconnectTime.strftime(r'%Y/%m/%d %H:%M:%S') if disconnectTime.year > 1601 else 'None'
			userName = self.sessions[i]['Domain'] + '\\' + self.sessions[i]['Username'] if len(self.sessions[i]['Username']) else ''

			session_entry = {
				"attributes": {
					"SessionName": self.sessions[i]['SessionName'],
					"Username": userName,
					"ID": str(i),
					"State": self.sessions[i]['state'],
					"Desktop": desktop_states[self.sessions[i]['flags']],
					"ConnectTime": connectTime,
					"DisconnectTime": disconnectTime,
					"ClientName": self.sessions[i]['ClientName'],
					"RemoteAddress": self.sessions[i]['RemoteIp'],
					"Resolution": self.sessions[i]['Resolution'],
					"ClientTimeZone": self.sessions[i]['ClientTimeZone']
				}
			}
			result.append(session_entry)

		return result

	def lookupSids(self):
		# Slightly modified code from lookupsid.py
		try:
			stringbinding = r'ncacn_np:%s[\pipe\lsarpc]' % self.__target_ip
			rpctransport = transport.DCERPCTransportFactory(stringbinding)
			rpctransport.set_smb_connection(self.__smbConnection)
			dce = rpctransport.get_dce_rpc()
			if self.__doKerberos:
				dce.set_auth_type(RPC_C_AUTHN_GSS_NEGOTIATE)
			dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
			dce.connect()

			dce.bind(lsat.MSRPC_UUID_LSAT)
			sids = list(self.sids.keys())
			if len(sids) > 32:
				sids = sids[:32] # TODO in future update
			resp = lsad.hLsarOpenPolicy2(dce, MAXIMUM_ALLOWED | lsat.POLICY_LOOKUP_NAMES)
			policyHandle = resp['PolicyHandle']
			try:
				resp = lsat.hLsarLookupSids(dce, policyHandle, sids, lsat.LSAP_LOOKUP_LEVEL.LsapLookupWksta)
			except DCERPCException as e:
				if str(e).find('STATUS_SOME_NOT_MAPPED') >= 0:
					resp = e.get_packet()
				else: 
					raise
			for sid, item in zip(sids,resp['TranslatedNames']['Names']):
				# if item['Use'] != SID_NAME_USE.SidTypeUnknown:
				domainIndex = item['DomainIndex']
				if domainIndex == -1: # Unknown domain
					self.sids[sid] = '{}\\{}'.format('???', item['Name'])
				elif domainIndex >= 0:
					name = '{}\\{}'.format(resp['ReferencedDomains']['Domains'][item['DomainIndex']]['Name'], item['Name'])
					self.sids[sid] = name
			dce.disconnect()
		except:
			logging.debug(traceback.format_exc())

	def sidToUser(self, sid):
		if sid[:2] == 'S-' and sid in self.sids:
			return self.sids[sid]
		return sid

	def do_tasklist(self, pid=None, name=None):
		with TSTS.LegacyAPI(self.__smbConnection, self.__target_ip, self.__doKerberos) as legacy:
			handle = legacy.hRpcWinStationOpenServer()
			r = legacy.hRpcWinStationGetAllProcesses(handle)
			if not len(r):
				return None

			self.sids = {}
			for procInfo in r:
				sid = procInfo['pSid']
				if sid[:2] == 'S-' and sid not in self.sids:
					self.sids[sid] = sid
			
			self.lookupSids()

			results = []
			self.get_session_list()
			self.enumerate_sessions_config()
			
			for procInfo in r:
				if pid is not None and pid not in str(procInfo['UniqueProcessId']):
					continue
				if name is not None and name.lower() not in procInfo['ImageName'].lower():
					continue
				
				sessId = procInfo['SessionId']
				fullUserName = ''
				if len(self.sessions[sessId]['Domain']):
					fullUserName += self.sessions[sessId]['Domain'] + '\\'
				if len(self.sessions[sessId]['Username']):
					fullUserName += self.sessions[sessId]['Username']
				
				entry = {
					"attributes": {
						"ImageName": procInfo['ImageName'],
						"PID": procInfo['UniqueProcessId'],
						"SessionID": procInfo['SessionId'],
						"SessionName": self.sessions[sessId]['SessionName'],
						"State": self.sessions[sessId]['state'],
						"SessionUser": fullUserName,
						"SID": self.sidToUser(procInfo['pSid']),
						"MemUsage": f"{procInfo['WorkingSetSize']//1000:,} K"
					}
				}
				results.append(entry)
			
			return results

	def do_taskkill(self, pid=None, name=None):
		if pid is None and name is None:
			logging.error('One of the following is required: -pid, -name')
			return
		if pid is not None and name is not None:
			logging.error('Only one of the following is required: -pid, -name')
			return

		pidList = []
		with TSTS.LegacyAPI(self.__smbConnection, self.__target_ip, self.__doKerberos) as legacy:
			handle = legacy.hRpcWinStationOpenServer()
			if pid is None and name is not None:
				r = legacy.hRpcWinStationGetAllProcesses(handle)
				if not len(r):
					logging.error('Could not get process list')
					return
				pidList = [i['UniqueProcessId'] for i in r if name.lower() in i['ImageName'].lower()]
				if not len(pidList):
					logging.error('Could not find %r in process list' % name)
					return
			else:
				pidList = [pid]

			for pid in pidList:
				logging.warning(f'Terminating PID: {pid} ...')
				try:
					if legacy.hRpcWinStationTerminateProcess(handle, pid)['ErrorCode']:
						return True
					else:
						return False
				except Exception as e:
					logging.error(f'Error terminating pid: {pid}')
					logging.error(str(e))
					return False

	def do_tscon(self):
		options = self.__options
		with TSTS.TermSrvSession(self.__smbConnection, self.__target_ip, self.__doKerberos) as TSSession:
			try:
				session_handle = None
				logging.debug('Connecting SessionID %d to %d ...' % (options.source, options.dest))
				try:
					session_handle = TSSession.hRpcOpenSession(options.source)
				except Exception as e:
					return False
					if e.error_code == 0x80070002:
						logging.error('Could not find source SessionID: %d' % options.source)
					else:
						logging.error(str(e))
					return False
				if TSSession.hRpcConnect(hSession = session_handle,
										 TargetSessionId = options.dest,
										 Password = options.password)['ErrorCode'] == 0:
					return True
				else:
					return False
			except Exception as e:
				return False
				if e.error_code == 0x80070002:
					logging.error('Could not find destination SessionID: %d' % options.dest)
				elif e.error_code == 0x8007139f:
					logging.error('Session in the invalid state. Did you mean %d -> %d?' % (options.dest, options.source))
				else:
					logging.error(str(e))

	def do_tsdiscon(self, session_id):
		with TSTS.TermSrvSession(self.__smbConnection, self.__target_ip, self.__doKerberos) as TSSession:
			try:
				logging.debug('Disconnecting SessionID: %d ...' % session_id)
				session_handle = TSSession.hRpcOpenSession(session_id)
				if TSSession.hRpcDisconnect(session_handle)['ErrorCode'] == 0:
					return True
				else:
					return False
			except Exception as e:
				return False
				if e.error_code == 1:
					logging.error('Maybe it is already disconnected?')
				elif e.error_code == 0x80070002:
					logging.error('Could not find SessionID: %d' % session_id)
				else:
					logging.error(str(e))

	def do_logoff(self, session_id):
		with TSTS.TermSrvSession(self.__smbConnection, self.__target_ip, self.__doKerberos) as TSSession:
			try:
				logging.warning('Signing-out SessionID: %d ...' % session_id)
				session_handle = TSSession.hRpcOpenSession(session_id)
				
				if TSSession.hRpcLogoff(session_handle)['ErrorCode'] == 0:
					return True
				else:
					return False
			except Exception as e:
				if e.error_code == 0x10000000:
					return True
				return False
				if e.error_code == 0x80070002:
					logging.error('Could not find SessionID: %d' % session_id)
				else:
					logging.error(str(e))

	def do_shutdown(self, logoff=False, shutdown=False, reboot=False, poweroff=False):
		if not logoff and not shutdown and not reboot and not poweroff:
			logging.error('No shutdown flags provided')
			return
		
		with TSTS.LegacyAPI(self.__smbConnection, self.__target_ip, self.__doKerberos) as legacy:
			handle = legacy.hRpcWinStationOpenServer()
			flags = 0
			flagsList = []
			shutdown_flags = [logoff, shutdown, reboot, poweroff]
			for k,v in zip(shutdown_flags, ['logoff', 'shutdown', 'reboot', 'poweroff']):
				if k:
					flagsList.append(v)
			flagsList = '|'.join(flagsList)
			for k,v in zip(shutdown_flags, [1,2,4,8]):
				if k:
					flags |= v
			try:
				logging.debug('Sending shutdown (%s) event ...' % (flagsList))
				resp = legacy.hRpcWinStationShutdownSystem(handle, 0, flags)
				if resp['ErrorCode']:
					return True
				else:
					resp.dump()
					return False
			except Exception as e:
				return False
				logging.error(str(e))
	

	def do_msg(self, session_id, title, message):
		with TSTS.TermSrvSession(self.__smbConnection, self.__target_ip, self.__doKerberos) as TSSession:
			try:
				logging.debug(f'Sending message to SessionID: {session_id} ...')
				session_handle = TSSession.hRpcOpenSession(session_id)
				if TSSession.hRpcShowMessageBox(session_handle, title, message)['ErrorCode'] == 0:
					return True
				else:
					return False
			except Exception as e:
				return False
				if e.error_code == 0x80070002:
					logging.error(f'Could not find SessionID: {session_id}')
				else:
					logging.error(str(e))


#!/usr/bin/env python3
import logging
from impacket.ldap import ldaptypes
from impacket.uuid import bin_to_string, string_to_bin
from impacket.dcerpc.v5 import rrp
from impacket.smb3 import SessionError
from ldap3.protocol.formatters.formatters import format_sid
from ldap3.protocol.microsoft import security_descriptor_control
from ldap3 import SUBTREE, BASE, LEVEL
import socket

from powerview.utils.helpers import strip_entry

from powerview.utils.helpers import (
    is_admin_sid,
    filetime_to_str,
    get_user_sids,
    host2ip,
    get_random_num,
    get_random_hex,
)
from powerview.utils.constants import (
    ACTIVE_DIRECTORY_RIGHTS,
    CERTIFICATE_RIGHTS,
    CERTIFICATION_AUTHORITY_RIGHTS,
    MS_PKI_CERTIFICATE_NAME_FLAG,
    MS_PKI_ENROLLMENT_FLAG,
    OID_TO_STR_MAP,
    WELL_KNOWN_SIDS,
    EXTENDED_RIGHTS_NAME_MAP
)

INHERITED_ACE = 0x10

#stolen from https://github.com/ly4k/Certipy
class ActiveDirectorySecurity:
    RIGHTS_TYPE = ACTIVE_DIRECTORY_RIGHTS

    def __init__(
        self,
        security_descriptor: bytes,
    ):
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        sd.fromString(security_descriptor)
        self.sd = sd

        self.owner = format_sid(sd["OwnerSid"].getData())
        self.aces = {}

        aces = sd["Dacl"]["Data"]
        for ace in aces:
            sid = format_sid(ace["Ace"]["Sid"].getData())

            if sid not in self.aces:
                self.aces[sid] = {
                    "rights": self.RIGHTS_TYPE(0),
                    "extended_rights": [],
                    "inherited": ace["AceFlags"] & INHERITED_ACE == INHERITED_ACE,
                }

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                self.aces[sid]["rights"] |= self.RIGHTS_TYPE(ace["Ace"]["Mask"]["Mask"])

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                if ace["Ace"]["Flags"] == 2:
                    uuid = bin_to_string(ace["Ace"]["InheritedObjectType"]).lower()
                elif ace["Ace"]["Flags"] == 1:
                    uuid = bin_to_string(ace["Ace"]["ObjectType"]).lower()
                else:
                    continue

                self.aces[sid]["extended_rights"].append(uuid)

#stolen from https://github.com/ly4k/Certipy
class SecurityDescriptorParser:
    """Base class for parsing security descriptors."""

    RIGHTS_TYPE = None  # Must be defined by subclasses

    def __init__(self, security_descriptor: bytes):
        """
        Initialize a security descriptor parser.

        Args:
            security_descriptor: Binary representation of a security descriptor
        """
        if self.RIGHTS_TYPE is None:
            raise NotImplementedError("Subclasses must define RIGHTS_TYPE")

        # Parse the security descriptor
        self.sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        self.sd.fromString(security_descriptor)

        # Extract owner SID
        self.owner = format_sid(self.sd["OwnerSid"].getData())

        # Dictionary to store access control entries by SID
        self.aces: Dict[str, Dict[str, Any]] = {}

        # Parse the ACEs
        self._parse_aces()

    def _parse_aces(self) -> None:
        """Parse the access control entries from the security descriptor."""
        pass  # To be implemented by subclasses

class CertificateSecurity(ActiveDirectorySecurity):
    RIGHTS_TYPE = CERTIFICATE_RIGHTS

class CASecurity(SecurityDescriptorParser):
    RIGHTS_TYPE = CERTIFICATION_AUTHORITY_RIGHTS
    def _parse_aces(self) -> None:
        """
        Parse the access control entries from the security descriptor.

        CA security descriptors have a simpler structure than AD security descriptors.
        """
        aces = self.sd["Dacl"]["Data"]

        for ace in aces:
            sid = format_sid(ace["Ace"]["Sid"].getData())

            if sid not in self.aces:
                self.aces[sid] = {
                    "rights": self.RIGHTS_TYPE(0),
                    "extended_rights": [],  # CAs don't use extended rights, but keeping for consistency
                    "inherited": bool(ace["AceFlags"] & INHERITED_ACE),
                }

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                mask = self.RIGHTS_TYPE(ace["Ace"]["Mask"]["Mask"])
                self.aces[sid]["rights"] |= mask

class CAEnum:
    def __init__(self, powerview, check_all=False):
        self.powerview = powerview
        self.ldap_session = self.powerview.conn.ldap_session
        self.ldap_server = self.powerview.conn.ldap_server
        self.root_dn = self.powerview.root_dn
        self.configuration_dn = self.powerview.configuration_dn
        self.check_all = check_all
        
    def fetch_root_ca(self, properties=['*']):
        enroll_filter = "(objectclass=certificationAuthority)"
        ca_search_base = f"CN=Certification Authorities,CN=Public Key Services,CN=Services,CN=Configuration,{self.root_dn}"
        logging.debug(f'LDAP Base: {ca_search_base}')
        logging.debug(f'LDAP Filter: {enroll_filter}')
        
        return self.ldap_session.extend.standard.paged_search(ca_search_base, enroll_filter, attributes=list(properties), paged_size=1000, generator=True)

    def fetch_enrollment_services(self,
            identity=None,
            properties=[
                "cn",
                "name",
                "dNSHostName",
                "cACertificateDN",
                "cACertificate",
                "certificateTemplates",
                "objectGUID",
            ],
            searchbase=None,
            search_scope=SUBTREE,
            no_cache=False,
            no_vuln_check=False,
            raw=False,
            include_sd=False
        ):
        if identity:
            identity_filter = f"(|(cn={identity})(name={identity}))"
        else:
            identity_filter = ""
        enroll_filter = f"(&(objectClass=pKIEnrollmentService){identity_filter})"

        if not searchbase:
            searchbase = f"CN=Enrollment Services,CN=Public Key Services,CN=Services,{self.configuration_dn}"

        if include_sd:
            properties.append("nTSecurityDescriptor")
            controls = security_descriptor_control(sdflags=0x05)
        else:
            controls = None

        logging.debug(f"LDAP Base: {searchbase}")
        logging.debug(f"LDAP Filter: {enroll_filter}")
        entries = self.ldap_session.extend.standard.paged_search(
            searchbase,
            enroll_filter,
            attributes=list(properties),
            paged_size=1000,
            generator=True,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw,
            controls=controls
        )
        if self.check_all:
            for entry in entries:
                # Access Rights + owner
                if not entry["attributes"]["nTSecurityDescriptor"]:
                    continue

                entry_rrp = self.get_rrp_config(computer_name=entry["attributes"]["dNSHostName"], ca=entry["attributes"]["name"])
                aces = entry_rrp["aces"].aces

                # Access Rights + owner
                security = CASecurity(entry["attributes"]["nTSecurityDescriptor"])
                entry["attributes"]["Owner"] = self.powerview.convertfrom_sid(security.owner)
                entry["attributes"]["ManageCa"] = []
                entry["attributes"]["ManageCertificates"] = []
                entry["attributes"]["Enroll"] = []
                entry["attributes"]["Vulnerabilities"] = []
                
                for sid, rights in aces.items():
                    if CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CA in rights["rights"]:
                        entry["attributes"]["ManageCa"].append(self.powerview.convertfrom_sid(sid))
                    if CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CERTIFICATES in rights["rights"]:
                        entry["attributes"]["ManageCertificates"].append(self.powerview.convertfrom_sid(sid))
                    if CERTIFICATION_AUTHORITY_RIGHTS.ENROLL in rights["rights"]:
                        entry["attributes"]["Enroll"].append(self.powerview.convertfrom_sid(sid))
                
                # Active Policy
                entry["attributes"]["ActivePolicy"] = entry_rrp["active_policy"]

                # Disabled Extensions
                entry["attributes"]["DisabledExtensions"] = entry_rrp["disable_extension_list"]

                # Request Disposition
                entry["attributes"]["RequestDisposition"] = (
                    "Pending" if entry_rrp["request_disposition"] & 0x100 else "Issue"
                )

                # User Specified SAN
                entry["attributes"]["UserSpecifiedSAN"] = (
                    "Enabled" if (entry_rrp["edit_flags"] & 0x00040000) == 0x00040000 else "Disabled"
                )

                # Enforce Encryption Flag
                entry["attributes"]["EnforceEncryptionFlag"] = (
                    "Enabled" if (entry_rrp["interface_flags"] & 0x00000200) == 0x00000200 else "Disabled"
                )

                # Vulnerabilities
                # ESC16
                if ("1.3.6.1.4.1.311.25.2" in entry["attributes"]["DisabledExtensions"]) and (entry["attributes"]["RequestDisposition"] in ["Issue", "Unknown"]):
                    # Check if current user is part of entry["attributes"]["Enroll"]
                    for sid in get_user_sids(self.powerview.get_domain_sid(), self.powerview.current_user_sid, self.ldap_session):
                        if entry["attributes"]["Enroll"][0] in self.powerview.convertfrom_sid(sid):
                            entry["attributes"]["Vulnerabilities"].append("ESC16")  
                
        return entries

    def get_rrp_config(self,
		computer_name,
        ca,
		port=445
	):

        KNOWN_PROTOCOLS = {
            139: {'bindstr': r'ncacn_np:%s[\pipe\winreg]', 'set_host': True},
            445: {'bindstr': r'ncacn_np:%s[\pipe\winreg]', 'set_host': True},
        }

        entries = {}

        stringBinding = KNOWN_PROTOCOLS[port]['bindstr'] % computer_name
        computer_name = host2ip(computer_name, self.powerview.conn.nameserver, 3, True, use_system_ns=self.powerview.conn.use_system_ns)
        try:
            dce = self.powerview.conn.connectRPCTransport(host=computer_name, stringBindings=stringBinding)
        except SessionError as e:
            logging.warning(f"[CAEnum] Failed to connect to {computer_name}, retrying...")
            self.get_rrp_config(computer_name, ca)

        if not dce:
            logging.error("[CAEnum] Failed to connect to %s" % (computer_name))
            return False

        for _ in range(3):
            try:
                dce.connect()
                _ = dce.bind(rrp.MSRPC_UUID_RRP)
                break
            except Exception as e:
                logging.error("[CAEnum] Failed to bind to %s" % (computer_name))
                continue

        try:
            # Open Local Machine registry handle
            hklm = rrp.hOpenLocalMachine(dce)
            h_root_key = hklm["phKey"]

            # First retrieve active policy module information
            policy_key_path = (
                f"SYSTEM\\CurrentControlSet\\Services\\CertSvc\\Configuration\\{ca}\\"
                "PolicyModules"
            )
            policy_key = rrp.hBaseRegOpenKey(dce, h_root_key, policy_key_path)

            # Get active policy module name
            _, active_policy = rrp.hBaseRegQueryValue(
                dce, policy_key["phkResult"], "Active"
            )

            if not isinstance(active_policy, str):
                logging.warning(f"[CAEnum] Expected a string, got {type(active_policy)!r} for {ca}")
                logging.warning(f"[CAEnum] Falling back to default policy")
                active_policy = "CertificationAuthority_MicrosoftDefault.Policy"

            active_policy = active_policy.strip("\x00")

            # Open policy module configuration
            policy_key_path = (
                f"SYSTEM\\CurrentControlSet\\Services\\CertSvc\\Configuration\\{ca}\\"
                f"PolicyModules\\{active_policy}"
            )
            policy_key = rrp.hBaseRegOpenKey(dce, h_root_key, policy_key_path)

            # Retrieve edit flags (controls certificate request behavior)
            _, edit_flags = rrp.hBaseRegQueryValue(
                dce, policy_key["phkResult"], "EditFlags"
            )

            if not isinstance(edit_flags, int):
                logging.warning(f"[CAEnum] Expected an integer, got {type(edit_flags)!r} for {ca}")
                logging.warning(f"[CAEnum] Falling back to default edit flags")
                edit_flags = 0x00000000

            # Retrieve request disposition (auto-enrollment settings    )
            _, request_disposition = rrp.hBaseRegQueryValue(
                dce, policy_key["phkResult"], "RequestDisposition"
            )

            if not isinstance(request_disposition, int):
                logging.warning(f"[CAEnum] Expected an integer, got {type(request_disposition)!r} for {ca}")
                logging.warning(f"[CAEnum] Falling back to default request disposition")
                request_disposition = 0x00000000

            
            # Retrieve disabled extensions
            _, disabled_extensions = rrp.hBaseRegQueryValue(
                dce, policy_key["phkResult"], "DisableExtensionList"
            )

            if not isinstance(disabled_extensions, str):
                logging.warning(f"[CAEnum] Expected a string, got {type(disabled_extensions)!r} for {ca}")
                logging.warning(f"[CAEnum] Falling back to default disabled extensions")
                disabled_extensions = ""

            # Process null-terminated string list into Python list
            disable_extension_list = [
                item for item in disabled_extensions.strip("\x00").split("\x00") if item
            ]

            # Now get general CA configuration settings
            configuration_key_path = (
                f"SYSTEM\\CurrentControlSet\\Services\\CertSvc\\Configuration\\{ca}\\"
            )
            configuration_key = rrp.hBaseRegOpenKey(dce, h_root_key, configuration_key_path)

            # Retrieve interface flags (controls CA interface behavior)
            _, interface_flags = rrp.hBaseRegQueryValue(
                dce, configuration_key["phkResult"], "InterfaceFlags"
            )

            if not isinstance(interface_flags, int):
                logging.warning(f"[CAEnum] Expected an integer, got {type(interface_flags)!r} for {ca}")
                logging.warning(f"[CAEnum] Falling back to default interface flags")
                interface_flags = 0x00000000

            # Retrieve security descriptor (controls access permissions)
            _, security_descriptor = rrp.hBaseRegQueryValue(
                dce, configuration_key["phkResult"], "Security"
            )

            if not isinstance(security_descriptor, bytes):
                logging.warning(f"[CAEnum] Expected bytes, got {type(security_descriptor)!r} for {ca}")
            
            security_descriptor = CASecurity(security_descriptor)

            entries = {
                "active_policy": active_policy,
                "edit_flags": edit_flags,
                "request_disposition": request_disposition,
                "disable_extension_list": disable_extension_list,
                "interface_flags": interface_flags,
                "aces": security_descriptor
            }
        except Exception as e:
            raise ValueError("[CAEnum] %s" % (str(e)))

        dce.disconnect()

        return entries

    def get_certificate_templates(self, properties=None, ca_search_base=None, identity=None, search_scope=SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        if not properties:
            properties=[
                "objectGUID",
                "cn",
                "name",
                "displayName",
                "pKIExpirationPeriod",
                "pKIOverlapPeriod",
                "msPKI-Enrollment-Flag",
                "msPKI-Private-Key-Flag",
                "msPKI-Certificate-Name-Flag",
                "msPKI-Certificate-Policy",
                "msPKI-Minimal-Key-Size",
                "msPKI-RA-Signature",
                "pKIExtendedKeyUsage",
                "nTSecurityDescriptor",
                "objectGUID",
                "whenCreated",
                "whenChanged",
                "msPKI-Template-Schema-Version"
            ]

        if not ca_search_base:
            ca_search_base = f"CN=Certificate Templates,CN=Public Key Services,CN=Services,{self.configuration_dn}"
        
        search_filter = ""
        identity_filter = ""

        if identity:
            identity_filter = f"(|(cn={identity})(displayName={identity}))"

        search_filter = f"(&(objectclass=pKICertificateTemplate){identity_filter})"

        logging.debug(f"LDAP Base: {ca_search_base}")
        logging.debug(f"LDAP Filter: {search_filter}")
        return self.ldap_session.extend.standard.paged_search(
            self.configuration_dn,
            search_filter,
            attributes=properties,
            controls=security_descriptor_control(sdflags=0x05),
            paged_size=1000,
            generator=True,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )

    # https://github.com/ly4k/Certipy/blob/main/certipy/commands/find.py#L688
    def check_web_enrollment(self, target, timeout=5):
        if target is None:
            logging.debug("No target found")
            return [False, False]  # [HTTP, HTTPS]

        results = [False, False]  # [HTTP, HTTPS]
        ports = [80, 443]
        protocols = ["HTTP", "HTTPS"]

        for i, port in enumerate(ports):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                logging.debug(f"Default timeout is set to {timeout}")
                sock.settimeout(timeout)
                logging.debug(f"Connecting to {target}:{port}")
                sock.connect((target, port))
                
                if port == 80:
                    sock.sendall(
                        "\r\n".join(
                            ["HEAD /certsrv/ HTTP/1.1", f"Host: {target}", "\r\n"]
                        ).encode()
                    )
                else:  # HTTPS
                    sock.sendall(
                        "\r\n".join(
                            ["HEAD /certsrv/ HTTP/1.1", f"Host: {target}", "Connection: close", "\r\n"]
                        ).encode()
                    )
                
                resp = sock.recv(256)
                sock.close()
                head = resp.split(b"\r\n")[0].decode()

                results[i] = " 404 " not in head
                logging.debug(f"{protocols[i]} enrollment check result: {results[i]}")
                
            except ConnectionRefusedError:
                logging.debug(f"Connection refused for {protocols[i]} on {target}:{port}")
                continue
            except socket.timeout:
                logging.debug(f"Can't reach {target}:{port} ({protocols[i]})")
                continue
            except Exception as e:
                logging.warning(
                    f"Got error while trying to check for {protocols[i]} enrollment: {e}"
                )
                continue
        
        if results[0] and results[1]:
            return [f"http://{target}/certsrv", f"https://{target}/certsrv"]
        elif results[0]:
            return [f"http://{target}/certsrv"]
        elif results[1]:
            return [f"https://{target}/certsrv"]
        else:
            return results

    def get_issuance_policies(self, properties=None, sdflags=0x5, no_cache=False, no_vuln_check=False, raw=False):
        if not properties:
            properties = [
                "cn",
                "name", 
                "displayName",
                "msDS-OIDToGroupLink",
                "msPKI-Cert-Template-OID",
                "nTSecurityDescriptor",
                "objectGUID"
            ]
        searchbase = f"CN=OID,CN=Public Key Services,CN=Services,{self.configuration_dn}"
        ldap_filter = "(objectclass=msPKI-Enterprise-Oid)"
        entries = []
        return self.ldap_session.extend.standard.paged_search(
            searchbase,
            ldap_filter,
            attributes=list(properties), 
            paged_size=1000,
            generator=True,
            controls=security_descriptor_control(sdflags=sdflags),
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def add_oid(self, template_name, template_oid, displayname=None, flags=0x01):
        oa = {
            'Name': template_name,
            'DisplayName': displayname if displayname else template_name,
            'flags': flags,
            'msPKI-Cert-Template-OID': template_oid,
        }
        oidpath = f"CN={template_oid},CN=OID,CN=Public Key Services,CN=Services,{self.configuration_dn}"
        self.ldap_session.add(oidpath, ['top','msPKI-Enterprise-Oid'], oa)
        if self.ldap_session.result['result'] == 0:
            logging.debug(f"[Add-DomainCATemplate] Added new template OID {oidpath}")
            logging.debug(f"[Add-DomainCATemplate] msPKI-Cert-Template-OID: {template_oid}")
            return True
        else:
            logging.error(f"[Add-DomainCATemplate] Error adding new template OID ({self.ldap_session.result['description']})")
            return False


class PARSE_TEMPLATE:
    def __init__(self, template, current_user_sid=None, linked_group=None, ldap_session=None):
        self.template = template
        self.owner_sid = None
        self.parsed_dacl = {}
        self.certificate_name_flag = None
        self.enrollment_flag = None
        self.extended_key_usage = None
        self.validity_period = None
        self.client_authentication = False
        self.enrollee_supplies_subject = False
        self.any_purpose = False
        self.enrollment_agent = False
        self.requires_manager_approval = False
        self.authorized_signatures_required = False
        self.no_security_extension = False
        self.domain_sid = None
        self.current_user_sid = current_user_sid
        self.linked_group = linked_group
        self.ldap_session = ldap_session

    def get_owner_sid(self):
        return self.owner_sid

    def set_owner_sid(self, owner_sid):
        self.owner_sid = owner_sid

    def get_parsed_dacl(self):
        return self.parsed_dacl

    def set_parsed_dacl(self, parsed_dacl):
        self.parsed_dacl = parsed_dacl

    def get_certificate_name_flag(self):
        return self.certificate_name_flag

    def set_certificate_name_flag(self, certificate_name_flag):
        self.certificate_name_flag = certificate_name_flag

    def get_enrollment_flag(self):
        return self.enrollment_flag

    def set_enrollment_flag(self, enrollment_flag):
        self.enrollment_flag = enrollment_flag

    def get_extended_key_usage(self):
        return self.extended_key_usage

    def set_extended_key_usage(self, extended_key_usage):
        self.extended_key_usage = extended_key_usage

    def get_validity_period(self):
        return self.validity_period

    def set_validity_period(self, validity_period):
        self.validity_period = validity_period

    def get_renewal_period(self):
        return self.renewal_period

    def set_renewal_period(self, renewal_period):
        self.renewal_period = renewal_period

    def get_client_authentication(self):
        return self.client_authentication

    def set_client_authentication(self, client_authentication):
        self.client_authentication = client_authentication

    def get_enrollee_supplies_subject(self):
        return self.enrollee_supplies_subject

    def set_enrollee_supplies_subject(self, enrollee_supplies_subject):
        self.enrollee_supplies_subject = enrollee_supplies_subject

    def get_any_purpose(self):
        return self.any_purpose

    def set_any_purpose(self, any_purpose):
        self.any_purpose = any_purpose

    def get_enrollment_agent(self):
        return self.enrollment_agent

    def set_enrollment_agent(self, enrollment_agent):
        self.enrollment_agent = enrollment_agent

    def set_requires_manager_approval(self, requires_manager_approval):
        self.requires_manager_approval = requires_manager_approval

    def get_requires_manager_approval(self):
        return self.requires_manager_approval

    def get_authorized_signatures_required(self):
        return self.authorized_signatures_required

    def set_authorized_signatures_required(self, authorized_signatures_required):
        self.authorized_signatures_required = authorized_signatures_required

    def get_no_security_extension(self):
        return self.no_security_extension

    def set_no_security_extension(self, no_security_extension):
        self.no_security_extension = no_security_extension

    def resolve_flags(self):
        # resolve certificate name flag
        self.set_certificate_name_flag(self.template.get("msPKI-Certificate-Name-Flag"))
        if self.certificate_name_flag is not None:
                self.set_certificate_name_flag(MS_PKI_CERTIFICATE_NAME_FLAG(
                    int(self.certificate_name_flag)
                ))
        else:
            self.set_certificate_name_flag(MS_PKI_CERTIFICATE_NAME_FLAG(0))

        # resolve enrollment flag
        self.set_enrollment_flag(self.template.get("msPKI-Enrollment-Flag"))
        if self.enrollment_flag is not None:
            self.set_enrollment_flag(MS_PKI_ENROLLMENT_FLAG(int(self.enrollment_flag)))
        else:
            self.set_enrollment_flag(MS_PKI_ENROLLMENT_FLAG(0))

        # resolve authorized signature
        self.set_authorized_signatures_required(self.template.get("msPKI-RA-Signature"))
        if self.authorized_signatures_required is not None:
                self.set_authorized_signatures_required(int(self.authorized_signatures_required))
        else:
            self.set_authorized_signatures_required(0)

        # resolve no_security_extension
        self.set_no_security_extension((
                MS_PKI_ENROLLMENT_FLAG.NO_SECURITY_EXTENSION in self.enrollment_flag
            ))

        # resolve pKIExtendedKeyUsage
        eku = self.template.get("pKIExtendedKeyUsage")
        if not isinstance(eku, list):
            if eku is None:
                eku = []
            else:
                eku = [eku]

        eku = list(map(lambda x: x.decode() if isinstance(x, bytes) else x, eku))
        self.set_extended_key_usage(list(
                map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, eku)
            ))

        self.set_any_purpose((
                "Any Purpose" in self.extended_key_usage or len(self.extended_key_usage) == 0
            ))

        self.set_client_authentication(self.any_purpose or any(
                eku in self.extended_key_usage
                for eku in [
                    "Client Authentication",
                    "Smart Card Logon",
                    "PKINIT Client Authentication",
                ]
            ))

        self.set_enrollment_agent(self.any_purpose or any(
                eku in self.extended_key_usage
                for eku in [
                    "Certificate Request Agent",
                ]
            ))

        self.set_enrollee_supplies_subject(any(
                flag in self.certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.ENROLLEE_SUPPLIES_SUBJECT,
                ]
            ))

        self.set_requires_manager_approval(
                MS_PKI_ENROLLMENT_FLAG.PEND_ALL_REQUESTS in self.enrollment_flag
        )

        # resolve validity period
        self.set_validity_period(self.template.get("pKIExpirationPeriod"))

        # resolve renewal_period
        self.set_renewal_period(self.template.get("pKIOverlapPeriod"))

    def can_user_enroll_template(self):
        enrollable_sids = []
        user_can_enroll = False
        for sid in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
            if sid in self.parsed_dacl["Enrollment Rights"]:
                enrollable_sids.append(sid)
                user_can_enroll = True
        return user_can_enroll, enrollable_sids

    def check_vulnerable_template(self):
        vulns = {}
        user_can_enroll, enrollable_sids = self.can_user_enroll_template()
        if not self.get_requires_manager_approval() and not self.get_authorized_signatures_required():
            # ESC1
            # TODO: add another user_can_enroll logic
            if (user_can_enroll and self.get_enrollee_supplies_subject() and self.get_client_authentication()):
                vulns["ESC1"] = enrollable_sids

            # ESC2
            if user_can_enroll and self.get_any_purpose():
                vulns["ESC2"] = enrollable_sids

            # ESC3
            if user_can_enroll and self.get_enrollment_agent():
                vulns["ESC3"] = enrollable_sids

            # ESC9
            if user_can_enroll and self.get_no_security_extension():
                vulns["ESC9"] = enrollable_sids

            # ESC13
            if user_can_enroll and self.get_client_authentication() and self.template["msPKI-Certificate-Policy"] and self.linked_group:
                vulns["ESC13"] = enrollable_sids

            # ESC15
            if user_can_enroll and self.get_enrollee_supplies_subject() and int(self.template.get("msPKI-Template-Schema-Version")) == 1:
                vulns["ESC15"] = enrollable_sids
        

        # ESC4
        security = CertificateSecurity(self.template.get("nTSecurityDescriptor"))
        owner_sid = security.owner

        if owner_sid in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
            vulns["ESC4"] = [owner_sid]
        else:
            has_vulnerable_acl = False
            aces = security.aces
            vulnerable_acl_sids = set()
            
            for sid, rights in aces.items():
                if sid not in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
                    continue

                ad_rights = rights["rights"] 
                ad_extended_rights = rights["extended_rights"]
                
                for right in [CERTIFICATE_RIGHTS.GENERIC_ALL, CERTIFICATE_RIGHTS.WRITE_OWNER, 
                             CERTIFICATE_RIGHTS.WRITE_DACL, CERTIFICATE_RIGHTS.WRITE_PROPERTY]:
                    if right in ad_rights:
                        vulnerable_acl_sids.add(sid)
                        has_vulnerable_acl = True
                
                if (CERTIFICATE_RIGHTS.WRITE_PROPERTY in ad_rights and
                    ('00000000-0000-0000-0000-000000000000' in ad_extended_rights and 
                     ad_rights & ACTIVE_DIRECTORY_RIGHTS.EXTENDED_RIGHT)):
                    vulnerable_acl_sids.add(sid)
                    has_vulnerable_acl = True

            if has_vulnerable_acl:
                vulns["ESC4"] = list(vulnerable_acl_sids)

        return vulns

    def create_object_ace(self,privguid,sid, mask=983551):
        nace = ldaptypes.ACE()
        nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
        nace['AceFlags'] = 0x02 # inherit to child objects
        acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
        acedata['Mask'] = ldaptypes.ACCESS_MASK()
        acedata['Mask']['Mask'] = mask # Full control
        acedata['ObjectType'] = string_to_bin(privguid)
        acedata['InheritedObjectType'] = b''
        acedata['Sid'] = ldaptypes.LDAP_SID()
        acedata['Sid'].fromCanonical(sid)
        assert sid == acedata['Sid'].formatCanonical()
        acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
        nace['Ace'] = acedata
        return nace
        
    def modify_dacl(self, sid, right_opt):
        permissions = {
                'all': {
                        'rights': [EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"], EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"]],
                        'mask': CERTIFICATE_RIGHTS.GENERIC_ALL,
                    },
                'enroll': {
                        'rights':[EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"]],
                        'mask': CERTIFICATE_RIGHTS.GENERIC_ALL,
                    },
                'write': {
                        'rights':[EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"]],
                        'mask':CERTIFICATE_RIGHTS.GENERIC_ALL,
                    }
                }
        sdData = self.template.get("nTSecurityDescriptor")
        if not sdData:
            raise Exception(f"No nTSecurityDescriptor found for template {self.template.get('cn')}")
        security = CertificateSecurity(sdData)
        for guid in permissions.get(right_opt).get('rights'):
            security.sd['Dacl']['Data'].append(self.create_object_ace(guid, sid, mask=permissions.get(right_opt).get('mask')))
        return security.sd.getData() 

    def parse_dacl(self):
        user_can_enroll = False
        enrollment_rights = []
        all_extended_rights = []

        sdData = self.template.get("nTSecurityDescriptor")
        if not sdData:
            raise Exception(f"No nTSecurityDescriptor found for template {self.template.get('cn')}")
        security = CertificateSecurity(sdData)
        self.owner_sid = security.owner
        if not self.domain_sid:
            self.domain_sid = '-'.join(self.owner_sid.split("-")[:-1])
        aces = security.aces

        for sid, rights in aces.items():
            # TODO: Fix Logic here
            # if sid in list(WELL_KNOWN_SIDS.keys()):
            #     continue

            if(EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
               or EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
               or EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"] in rights["extended_rights"]
               or CERTIFICATE_RIGHTS.GENERIC_ALL in rights["rights"]
               ):
                enrollment_rights.append(sid)

            if (
                EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"] in rights["extended_rights"]
            ):
                all_extended_rights.append(sid)

        rights_mapping = [
            (CERTIFICATE_RIGHTS.GENERIC_ALL, [], "Full Control Principals"),
            (CERTIFICATE_RIGHTS.WRITE_OWNER, [], "Write Owner Principals"),
            (CERTIFICATE_RIGHTS.WRITE_DACL, [], "Write Dacl Principals"),
            (
                CERTIFICATE_RIGHTS.WRITE_PROPERTY,
                [],
                "Write Property Principals",
            ),
        ]
        object_control_permissions = {}
        # acl
        for sid, rights in security.aces.items():
            rights = rights["rights"]

            for (right, principal_list, _) in rights_mapping:
                if right in rights:
                    principal_list.append(sid)

        for _, rights, name in rights_mapping:
            if len(rights) > 0:
                object_control_permissions[name] = rights

        self.parsed_dacl['Write Owner'] = object_control_permissions['Write Owner Principals']
        self.parsed_dacl['Write Dacl'] = object_control_permissions['Write Dacl Principals']
        self.parsed_dacl['Write Property'] = object_control_permissions['Write Property Principals']
        self.parsed_dacl['Enrollment Rights'] = enrollment_rights
        self.parsed_dacl['Extended Rights'] = all_extended_rights

        return self.parsed_dacl

class UTILS:
    @staticmethod
    def get_template_oid(oid_forest):
        oid_part_1 = get_random_num(10000000,99999999)
        oid_part_2 = get_random_num(10000000,99999999)
        oid_part_3 = get_random_hex(32)
        
        template_oid = f"{oid_forest}.{oid_part_1}.{oid_part_2}"
        templatename = f"{oid_part_2}.{oid_part_3}"

        return template_oid, templatename
        




from os import system

class bcolors:
	HEADER = '\033[95m'
	OKBLUE = '\033[94m'
	OKCYAN = '\033[96m'
	OKGREEN = '\033[92m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'
	BLACK = '\033[30m'
	RED = '\033[31m'
	GREEN = '\033[32m'
	YELLOW = '\033[33m'
	BLUE = '\033[34m'
	MAGENTA = '\033[35m'
	CYAN = '\033[36m'
	WHITE = '\033[37m'
	BRIGHTBLACK = '\033[90m'
	BRIGHTRED = '\033[91m'
	BRIGHTGREEN = '\033[92m'
	BRIGHTYELLOW = '\033[93m'
	BRIGHTBLUE = '\033[94m'
	BRIGHTMAGENTA = '\033[95m'
	BRIGHTCYAN = '\033[96m'
	BRIGHTWHITE = '\033[97m'
	BG_BLACK = '\033[40m'
	BG_RED = '\033[41m'
	BG_GREEN = '\033[42m'
	BG_YELLOW = '\033[43m'
	BG_BLUE = '\033[44m'
	BG_MAGENTA = '\033[45m'
	BG_CYAN = '\033[46m'
	BG_WHITE = '\033[47m'
	ITALIC = '\033[3m'
	STRIKETHROUGH = '\033[9m'
	INVERT = '\033[7m'
	GREY = '\033[2;37m'

class Gradient:
	"""
	Advanced text gradient effects for terminal output.
	"""
	
	@staticmethod
	def fire(text):
		"""Red to yellow gradient."""
		system(""); faded = ""
		green = 250
		for line in text.splitlines():
			faded += (f"\033[38;2;255;{green};0m{line}\033[0m\n")
			if not green == 0:
				green -= 25
				if green < 0:
					green = 0
		return faded

	@staticmethod
	def water(text):
		"""Blue to cyan gradient."""
		system(""); faded = ""
		green = 10
		for line in text.splitlines():
			faded += (f"\033[38;2;0;{green};255m{line}\033[0m\n")
			if not green == 255:
				green += 15
				if green > 255:
					green = 255
		return faded
	
	@staticmethod
	def purple(text):
		"""Purple to pink gradient."""
		system(""); faded = ""
		blue = 255
		for line in text.splitlines():
			faded += (f"\033[38;2;128;0;{blue}m{line}\033[0m\n")
			if not blue == 0:
				blue -= 25
				if blue < 0:
					blue = 0
		return faded
	
	@staticmethod
	def forest(text):
		"""Dark green to light green gradient."""
		system(""); faded = ""
		green = 100
		for line in text.splitlines():
			faded += (f"\033[38;2;0;{green};0m{line}\033[0m\n")
			if not green == 255:
				green += 15
				if green > 255:
					green = 255
		return faded
	
	@staticmethod
	def sunset(text):
		"""Orange to purple gradient."""
		system(""); faded = ""
		red = 255
		blue = 0
		for line in text.splitlines():
			faded += (f"\033[38;2;{red};100;{blue}m{line}\033[0m\n")
			if not red == 128:
				red -= 12
				blue += 12
				if red < 128:
					red = 128
				if blue > 128:
					blue = 128
		return faded
	
	@staticmethod
	def horizontal_rainbow(text, width=80):
		"""Creates a rainbow gradient across a single line of text."""
		system("")
		if not text:
			return ""
		
		result = ""
		text_length = len(text)
		for i, char in enumerate(text):
			# Calculate hue based on position (0-360 degrees in HSV color space)
			hue = int(360 * (i / width) % 360)
			
			# Convert HSV to RGB with full saturation and value
			h = hue / 60
			c = 255  # Full color intensity
			x = int(c * (1 - abs(h % 2 - 1)))
			
			if 0 <= h < 1: r, g, b = c, x, 0
			elif 1 <= h < 2: r, g, b = x, c, 0
			elif 2 <= h < 3: r, g, b = 0, c, x
			elif 3 <= h < 4: r, g, b = 0, x, c
			elif 4 <= h < 5: r, g, b = x, 0, c
			else: r, g, b = c, 0, x
			
			result += f"\033[38;2;{r};{g};{b}m{char}\033[0m"
		
		return result
		
	@staticmethod
	def generate_gradient_colors(start_rgb, end_rgb, steps):
		"""
		Generate a list of RGB color values forming a gradient.
		
		Args:
			start_rgb: Starting color as [r, g, b] with values 0-255
			end_rgb: Ending color as [r, g, b] with values 0-255
			steps: Number of colors to generate
			
		Returns:
			List of [r, g, b] values representing the gradient
		"""
		if steps <= 1:
			return [start_rgb]
			
		result = []
		for i in range(steps):
			ratio = i / (steps - 1)
			r = round(start_rgb[0] + (end_rgb[0] - start_rgb[0]) * ratio)
			g = round(start_rgb[1] + (end_rgb[1] - start_rgb[1]) * ratio)
			b = round(start_rgb[2] + (end_rgb[2] - start_rgb[2]) * ratio)
			result.append([r, g, b])
			
		return result